import collections
import copy
import json
import math
import os
import pickle
from functools import partial
from itertools import islice
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import femr
import femr.ontology
import numpy as np
import scipy.stats as stats
import transformers
from cehrbert.models.hf_models.tokenization_utils import agg_helper, load_json_file
from cehrbert.runners.hf_runner_argument_dataclass import DataTrainingArguments
from cehrbert_data.const.artificial_tokens import DEATH_TOKEN
from datasets import Dataset, DatasetDict, IterableDataset
from femr.stat_utils import OnlineStatistics, ReservoirSampler
from meds import death_code
from scipy.interpolate import UnivariateSpline
from tokenizers import AddedToken, Tokenizer
from tokenizers.models import WordLevel
from tokenizers.pre_tokenizers import WhitespaceSplit
from tokenizers.trainers import WordLevelTrainer
from tqdm import tqdm
from transformers import PreTrainedTokenizer
from transformers.utils import logging

from cehrgpt.gpt_utils import (
    convert_time_interval_to_time_tuple,
    extract_time_interval_in_days,
    is_att_token,
    is_clinical_event,
    is_inpatient_att_token,
)
from cehrgpt.models.pretrained_embeddings import PretrainedEmbeddings
from cehrgpt.models.special_tokens import (
    END_TOKEN,
    LINEAR_PROB_TOKEN,
    OUT_OF_VOCABULARY_TOKEN,
    PAD_TOKEN,
    START_TOKEN,
)
from cehrgpt.omop.ontology import Ontology

NUM_OF_BINS = 10
DEGREE_OF_FREEDOM = 3
SAMPLE_SIZE = 10_000
NA = "N/A"
UNKNOWN_BIN = "BIN:unknown"
NONE_BIN = "BIN:NONE"
TOKENIZER_FILE_NAME = "cehrgpt_tokenizer.json"
VALUE_TOKENIZER_FILE_NAME = "cehrgpt_value_tokenizer.json"
TIME_TOKENIZER_FILE_NAME = "cehrgpt_time_tokenizer.json"
TOKEN_TO_SUB_TIME_TOKEN_MAPPING_FILE_NAME = "token_to_sub_time_token_mapping.json"
LAB_STATS_FILE_NAME = "cehrgpt_lab_stats.pickle"
LEGACY_LAB_STATS_FILE_NAME = "cehrgpt_lab_stats.json"
CONCEPT_STATS_FILE_NAME = "cehrgpt_concept_stats.json"
DEMOGRAPHICS_STATS_FILE_NAME = "demographics_stats.pickle"
CONCEPT_MAPPING_FILE_NAME = "concept_name_mapping.json"
MOTOR_TIME_TO_EVENT_TASK_INFO_FILE_NAME = "motor_time_to_event_info.pickle"
ONTOLOGY_FILE_NAME = "ontology.pickle"

LOG = logging.get_logger("transformers")


def truncated_sample(sample, standard_deviation):
    lower_quantile = stats.norm.cdf(-standard_deviation)
    upper_quantile = stats.norm.cdf(standard_deviation)
    lower_bound = np.quantile(sample, lower_quantile)
    upper_bound = np.quantile(sample, upper_quantile)
    return [x for x in sample if lower_bound <= x <= upper_bound]


def is_valid_valid_bin(token: str) -> bool:
    return token.startswith("BIN:")


def create_value_bin(bin_index: int) -> str:
    return "BIN:" + str(bin_index)


def get_dataset_len(dataset: Union[Dataset, IterableDataset]) -> int:
    if isinstance(dataset, Dataset):
        return len(dataset)
    elif isinstance(dataset, IterableDataset):
        return sum([1 for _ in dataset])
    raise RuntimeError(
        "The dataset must be one of the two types (Dataset, IterableDataset)"
    )


def get_allowed_motor_codes(
    original_concept_codes: List[str], ontology: Optional[Ontology]
) -> List[str]:
    filtered_original_concept_codes = filter(is_clinical_event, original_concept_codes)
    if ontology:
        allowed_motor_codes = []
        for concept in filtered_original_concept_codes:
            domain = ontology.get_domain(concept)
            if domain and domain in ["Condition", "Procedure", "Drug", "Visit"]:
                allowed_motor_codes.append(concept)
            elif concept in [DEATH_TOKEN, death_code]:
                allowed_motor_codes.append(concept)
        return allowed_motor_codes
    else:
        return list(filtered_original_concept_codes)


def create_sample_from_bins(bins, sample_size: int = 10_000) -> List[float]:
    """
    Generates a specified number of samples from a list of bins, each containing a fitted spline.

    This function iterates over each bin, extracts the spline, and uses it to generate a set of samples
    uniformly distributed along the x-axis defined by the spline's knots. It ensures that the total number
    of samples generated matches the specified sample size by distributing the number of samples evenly
    across the bins.

    Parameters:
        bins (List[Dict[str, UnivariateSpline]]): A list of dictionaries, each containing a 'spline' key
            with a UnivariateSpline object as its value. These splines define the data distribution within
            each bin from which samples are to be generated.
        sample_size (int, optional): The total number of samples to generate from all bins combined.
            Defaults to 10,000.

    Returns:
        List[float]: A list of sampled values, where each value is generated based on the spline functions
            provided in the bins. The total number of samples in the list will be equal to `sample_size`.

    Raises:
        ValueError: If `sample_size` is less than the number of bins, as it would not be possible to generate
            at least one sample per bin.

    Example:
        >>> x = np.linspace(0, 10, 100)
        >>> y = np.sin(x)
        >>> spline = UnivariateSpline(x, y, s=1)
        >>> bins = [{'spline': spline} for _ in range(5)]
        >>> samples = create_sample_from_bins(bins, 1000)
        >>> len(samples)
        1000

    Note:
        The function assumes that each bin's spline has a sufficient range of x-values (knots) to allow for
        meaningful sampling. If the range of x-values is too narrow, the uniformity of the sample distribution
        may be affected.
    """
    sample = []
    num_of_bins = len(bins)
    if num_of_bins > 0:
        sample_per_bin = sample_size // num_of_bins
        for value_bin in bins:
            bin_spline = value_bin["spline"]
            x = np.random.uniform(
                bin_spline.get_knots()[0], bin_spline.get_knots()[-1], sample_per_bin
            )
            y = bin_spline(x)
            sample.extend(y)
    return sample


def create_bins_with_spline(samples, num_bins, d_freedom=3) -> List[Dict[str, Any]]:
    """
    Divides a list of numeric samples into a specified number of bins and fits a spline to the data in each bin.

    This function first sorts the list of samples, then partitions the sorted list into `num_bins` bins. For each bin,
    a UnivariateSpline is fitted to the data within the bin, using the specified degrees of freedom. The function
    handles edge cases by assigning infinity to the bounds of the first and last bins.

    Parameters:
        samples (List[float]): A list of sample data points, which are real numbers.
        num_bins (int): The number of bins to divide the sample data into. It is assumed that there are enough samples to at least fill the bins to the minimum required for spline fitting.
        d_freedom (int, optional): The degree of freedom for the spline. Default is 1, which fits a linear spline.

    Returns:
        List[Dict[str, Any]]: A list of dictionaries, each representing a bin. Each dictionary contains:
            - 'bin_index' (int): The index of the bin.
            - 'start_val' (float): The starting value of the bin, with the first bin starting at negative infinity.
            - 'end_val' (float): The ending value of the bin, with the last bin ending at positive infinity.
            - 'spline' (UnivariateSpline): The spline object fitted to the data within the bin.

    Raises:
        ValueError: If `num_bins` is less than 1 or if there are insufficient samples to create the specified number of bins with the required minimum data points per bin.

    Example:
        >>> samples = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        >>> bins = create_bins_with_spline(samples, 2)
        >>> for b in bins:
        ...     print(b['bin_index'], b['start_val'], b['end_val'])
        ...
        0 -inf 5.5
        1 5.5 inf

    Note:
        The function assumes that each bin will have at least `d_freedom + 1` samples to fit the spline. If the total number of samples is less than `num_bins * (d_freedom + 1)`, no bins will be created.
    """
    samples.sort()
    bins = []
    if len(samples) >= num_bins * (d_freedom + 1):
        samples_per_bin = len(samples) // num_bins
        for bin_index in range(0, num_bins):
            if bin_index == 0:
                start_val = float("-inf")
            else:
                start_val = samples[bin_index * samples_per_bin]

            if bin_index == num_bins - 1:
                end_val = float("inf")
            else:
                end_val = samples[(bin_index + 1) * samples_per_bin]
            x = range(bin_index * samples_per_bin, (bin_index + 1) * samples_per_bin)
            y = samples[bin_index * samples_per_bin : (bin_index + 1) * samples_per_bin]
            spline = UnivariateSpline(x, y, k=d_freedom)
            bins.append(
                {
                    "bin_index": bin_index,
                    "start_val": start_val,
                    "end_val": end_val,
                    "spline": spline,
                }
            )
    return bins


def map_motor_tte_statistics(
    batch: Dict[str, Any],
    allowed_motor_codes: List[str],
) -> Dict[str, Any]:
    motor_event_times = femr.stat_utils.ReservoirSampler(100_000)
    task_tte_stats: Dict[str, int] = collections.defaultdict(int)
    task_censor_stats: Dict[str, int] = collections.defaultdict(int)
    for concept_ids in batch["concept_ids"]:
        # First collect TTE data in reverse chronological order
        censor_time = 0
        time_to_event_dict: Dict[str, int] = {}
        next_future_visit_concepts = set()
        # Reverse walk through concept_ids to calculate TTE from each [VE] point
        for concept_id in reversed(concept_ids):
            if is_att_token(concept_id):
                time_interval = extract_time_interval_in_days(concept_id)
                if time_interval > 0:
                    # Update TTE for existing concepts, or add new ones seen in this visit
                    for existing_concept_id in list(time_to_event_dict.keys()):
                        if existing_concept_id in next_future_visit_concepts:
                            time_to_event_dict[existing_concept_id] = time_interval
                        else:
                            time_to_event_dict[existing_concept_id] += time_interval

                    for next_concept_id in next_future_visit_concepts:
                        if next_concept_id not in time_to_event_dict:
                            time_to_event_dict[next_concept_id] = time_interval

                    # Record the censor time at the end of the visit
                    censor_time += time_interval

                    # Keep track of the time to event value
                    for tte in time_to_event_dict.values():
                        motor_event_times.add(tte, 1)

                    for motor_code in allowed_motor_codes:
                        if motor_code in time_to_event_dict:
                            task_tte_stats[motor_code] += 1
                        else:
                            task_censor_stats[motor_code] += 1
                    next_future_visit_concepts.clear()
            else:
                next_future_visit_concepts.add(concept_id)

    return {
        "motor_event_times": motor_event_times,
        "task_tte_stats": task_tte_stats,
        "task_censor_stats": task_censor_stats,
    }


def compute_motor_tte_statistics(
    dataset: Dataset,
    data_args: DataTrainingArguments,
    allowed_motor_codes: List[str],
    ontology: Optional[Ontology] = None,
) -> Dict[str, Any]:
    map_motor_tte_statistics_partial = partial(
        map_motor_tte_statistics,
        allowed_motor_codes=allowed_motor_codes,
    )
    if data_args.streaming:
        first_example = next(iter(dataset))
        parts = dataset.map(
            partial(agg_helper, map_func=map_motor_tte_statistics_partial),
            batched=True,
            batch_size=data_args.preprocessing_batch_size,
            remove_columns=first_example.keys(),
        )
    else:
        parts = dataset.map(
            partial(agg_helper, map_func=map_motor_tte_statistics_partial),
            batched=True,
            batch_size=data_args.preprocessing_batch_size,
            remove_columns=dataset.column_names,
            num_proc=data_args.preprocessing_num_workers,
            keep_in_memory=True,
            new_fingerprint="invalid",
        )
    current = None
    for stat in tqdm(parts, desc="Aggregating the MOTOR TTE statistics"):
        fixed_stat = pickle.loads(stat["data"])
        if current is None:
            current = fixed_stat
        else:
            current["motor_event_times"].combine(fixed_stat["motor_event_times"])
            for k, v in fixed_stat["task_tte_stats"].items():
                current["task_tte_stats"][k] += v
            for k, v in fixed_stat["task_censor_stats"].items():
                current["task_censor_stats"][k] += v

    # Aggregate the counts for the parent concepts
    if ontology is not None:
        for k in list(current["task_tte_stats"].keys()):
            for parent in ontology.get_all_parents(k):
                if parent != k:
                    current["task_tte_stats"][parent] += current["task_tte_stats"][k]
        for k in list(current["task_censor_stats"].keys()):
            for parent in ontology.get_all_parents(k):
                if parent != k:
                    current["task_censor_stats"][parent] += current[
                        "task_censor_stats"
                    ][k]
    return current


def agg_statistics(stats1, stats2):
    if stats1.get("numeric_stats_by_lab"):
        for k, v in stats2["numeric_stats_by_lab"].items():
            stats1["numeric_stats_by_lab"][k].combine(v)
    if stats1.get("categorical_stats_by_lab"):
        for (concept_id, concept_as_value), count in stats2[
            "categorical_stats_by_lab"
        ].items():
            stats1["categorical_stats_by_lab"][(concept_id, concept_as_value)] += count
    if stats1.get("concept_code_stats"):
        for concept_id, weight in stats2["concept_code_stats"].items():
            stats1["concept_code_stats"][concept_id] += weight
    if stats1.get("gender_list"):
        stats1.get("gender_list").update(stats2.get("gender_list"))
    if stats1.get("race_list"):
        stats1.get("race_list").update(stats2.get("race_list"))
    return stats1


def map_statistics(batch: Dict[str, Any], total_size, size=10_000) -> Dict[str, Any]:
    if "units" in batch:
        batch_value_units = batch["units"]
    else:
        batch_value_units = [[NA for _ in cons] for cons in batch["concept_ids"]]

    if "number_as_values" not in batch:
        batched_number_as_values = [
            [value if isinstance(value, float) else None for value in concept_values]
            for concept_values in batch["concept_values"]
        ]
    else:
        batched_number_as_values = batch["number_as_values"]

    if "concept_as_values" not in batch:
        batched_concept_as_values = [
            [value if isinstance(value, str) else None for value in concept_values]
            for concept_values in batch["concept_values"]
        ]
    else:
        batched_concept_as_values = batch["concept_as_values"]

    numeric_stats_by_lab = collections.defaultdict(partial(ReservoirSampler, size=size))
    categorical_stats_by_lab = collections.defaultdict(int)
    concept_code_stats = collections.defaultdict(int)
    gender_list = set()
    race_list = set()
    for (
        concept_ids,
        number_as_values,
        concept_as_values,
        concept_value_indicators,
        units,
    ) in zip(
        batch["concept_ids"],
        batched_number_as_values,
        batched_concept_as_values,
        batch["concept_value_masks"],
        batch_value_units,
    ):
        # Collecting demographics
        gender, race = concept_ids[2:4]
        gender_list.add(gender)
        race_list.add(race)

        unique_codes = set()
        for (
            concept_id,
            number_as_value,
            concept_as_value,
            concept_value_indicator,
            unit,
        ) in zip(
            concept_ids,
            number_as_values,
            concept_as_values,
            concept_value_indicators,
            units,
        ):
            if concept_value_indicator == 1:
                if number_as_value:
                    numeric_stats_by_lab[(concept_id, unit)].add(number_as_value, 1)
                if concept_as_value:
                    categorical_stats_by_lab[(concept_id, concept_as_value)] += 1
            unique_codes.add(concept_id)

        for code in unique_codes:
            concept_code_stats[code] += 1 / total_size

    return {
        "numeric_stats_by_lab": numeric_stats_by_lab,
        "categorical_stats_by_lab": categorical_stats_by_lab,
        "concept_code_stats": concept_code_stats,
        "gender_list": gender_list,
        "race_list": race_list,
    }


def compute_statistics(
    dataset: Dataset,
    data_args: DataTrainingArguments,
    ontology: Optional[Ontology] = None,
) -> Dict[str, Any]:
    total = get_dataset_len(dataset)
    map_statistics_partial = partial(map_statistics, total_size=total, size=SAMPLE_SIZE)
    if data_args.streaming:
        first_example = next(iter(dataset))
        parts = dataset.map(
            partial(agg_helper, map_func=map_statistics_partial),
            batched=True,
            batch_size=data_args.preprocessing_batch_size,
            remove_columns=first_example.keys(),
        )
    else:
        parts = dataset.map(
            partial(agg_helper, map_func=map_statistics_partial),
            batched=True,
            batch_size=data_args.preprocessing_batch_size,
            remove_columns=dataset.column_names,
            num_proc=data_args.preprocessing_num_workers,
            keep_in_memory=True,
            new_fingerprint="invalid",
        )
    current = None
    for stat in tqdm(parts, desc="Aggregating the lab statistics"):
        fixed_stat = pickle.loads(stat["data"])
        if current is None:
            current = fixed_stat
        else:
            current = agg_statistics(current, fixed_stat)

    numeric_lab_stats = []
    for (concept_id, unit), online_stats in current["numeric_stats_by_lab"].items():
        if len(online_stats.samples) == 0:
            continue
        samples = truncated_sample(online_stats.samples, data_args.value_outlier_std)
        bins = create_bins_with_spline(samples, NUM_OF_BINS, DEGREE_OF_FREEDOM)
        if len(bins) > 0:
            numeric_lab_stats.append(
                {
                    "concept_id": concept_id,
                    "unit": unit,
                    "mean": np.mean(samples),
                    "std": np.std(samples),
                    "count": len(online_stats.samples),
                    "value_outlier_std": data_args.value_outlier_std,
                    "bins": bins,
                }
            )

    categorical_lab_stats = collections.defaultdict(int)
    for (concept_id, value_as_concept), count in current[
        "categorical_stats_by_lab"
    ].items():
        categorical_lab_stats[(concept_id, value_as_concept)] += count

    all_concept_code_stats = collections.defaultdict(float)
    for concept_id, count in current["concept_code_stats"].items():
        if ontology is not None:
            parents = ontology.get_all_parents(concept_id)
            for parent in parents:
                all_concept_code_stats[parent] += count
        else:
            all_concept_code_stats[concept_id] += count

    all_concept_code_entropies = collections.defaultdict(float)
    for concept_id, weight in all_concept_code_stats.items():
        baseline = (
            min(
                [1]
                + [
                    all_concept_code_stats[parent]
                    for parent in ontology.get_parents(concept_id)
                ]
            )
            if ontology is not None
            else 1
        )
        weight = weight / baseline
        weight = min(1.0, weight)
        if weight != 0 and weight != 1:
            weight = baseline * (
                weight * math.log(weight) + (1 - weight) * math.log(1 - weight)
            )
            all_concept_code_entropies[concept_id] = weight

    return {
        "numeric_lab_stats": numeric_lab_stats,
        "categorical_lab_stats": categorical_lab_stats,
        "original_concept_codes": list(current["concept_code_stats"].keys()),
        "all_concept_code_stats": all_concept_code_stats,
        "all_concept_code_entropies": all_concept_code_entropies,
        "gender_list": current["gender_list"],
        "race_list": current["race_list"],
        "total": total,
    }


def create_numeric_concept_unit_mapping(
    lab_stats: List[Dict[str, Any]]
) -> Tuple[Dict[str, List[float]], Dict[str, List[str]]]:
    numeric_concept_unit_mapping = collections.defaultdict(list)
    for each_lab_stat in lab_stats:
        numeric_concept_unit_mapping[each_lab_stat["concept_id"]].append(
            (each_lab_stat["count"], each_lab_stat["unit"])
        )

    concept_prob_mapping = dict()
    concept_unit_mapping = dict()
    for concept_id in numeric_concept_unit_mapping.keys():
        counts, units = zip(*numeric_concept_unit_mapping[concept_id])
        total_count = sum(counts)
        probs = [float(c) / total_count for c in counts]
        concept_prob_mapping[concept_id] = probs
        concept_unit_mapping[concept_id] = units
    return concept_prob_mapping, concept_unit_mapping


class NumericEventStatistics:
    def __init__(self, lab_stats: List[Dict[str, Any]]):
        self._lab_stats = lab_stats
        self._lab_stats_mapping = {
            (lab_stat["concept_id"], lab_stat["unit"]): {
                "unit": lab_stat["unit"],
                "mean": lab_stat["mean"],
                "std": lab_stat["std"],
                "value_outlier_std": lab_stat["value_outlier_std"],
                "bins": lab_stat["bins"],
            }
            for lab_stat in lab_stats
        }
        self._concept_prob_mapping, self._concept_unit_mapping = (
            create_numeric_concept_unit_mapping(lab_stats)
        )

    def get_numeric_concept_ids(self) -> List[str]:
        return [_["concept_id"] for _ in self._lab_stats]

    def get_random_unit(self, concept_id: str) -> str:
        if concept_id in self._concept_prob_mapping:
            unit_probs = self._concept_prob_mapping[concept_id]
            return np.random.choice(
                self._concept_unit_mapping[concept_id], p=unit_probs
            )
        return NA

    def normalize(
        self, concept_id: str, unit: str, concept_value: Union[float, str]
    ) -> str:
        if isinstance(concept_value, float):
            if (concept_id, unit) in self._lab_stats_mapping:
                concept_unit_stats = self._lab_stats_mapping[(concept_id, unit)]
                bins = concept_unit_stats["bins"]
                if bins:
                    for each_bin in bins:
                        if (
                            each_bin["start_val"]
                            <= concept_value
                            <= each_bin["end_val"]
                        ):
                            return create_value_bin(each_bin["bin_index"])
        return UNKNOWN_BIN

    def denormalize(
        self, concept_id: str, value_bin: str
    ) -> Tuple[Optional[Union[float, str]], str]:
        unit = self.get_random_unit(concept_id)
        concept_value = value_bin
        if (
            is_valid_valid_bin(value_bin)
            and (concept_id, unit) in self._lab_stats_mapping
        ):
            lab_stats = self._lab_stats_mapping[(concept_id, unit)]
            bin_index = value_bin.split(":")[1]
            if bin_index.isnumeric():
                bin_index = int(bin_index)
                # There are rare cases during sequence generation where bin_index could be out of range
                # when there are no bins for (concept_id, unit) due to the small number of values in the source data
                if len(lab_stats["bins"]) > bin_index:
                    assert bin_index == lab_stats["bins"][bin_index]["bin_index"]
                    bin_spline = lab_stats["bins"][bin_index]["spline"]
                    x = np.random.uniform(
                        bin_spline.get_knots()[0], bin_spline.get_knots()[-1]
                    )
                    concept_value = bin_spline(x).item()
        return concept_value, unit


class CehrGptTokenizer(PreTrainedTokenizer):

    def __init__(
        self,
        tokenizer: Tokenizer,
        value_tokenizer: Tokenizer,
        att_tokenizer: Tokenizer,
        token_to_sub_time_token_mapping: Dict[str, List[str]],
        concept_code_stats: Dict[str, Any],
        numeric_lab_stats: List[Dict[str, Any]],
        categorical_lab_stats: Dict[Tuple[str, str], int],
        concept_name_mapping: Dict[str, str],
        pretrained_concept_embedding_model: PretrainedEmbeddings = None,
        motor_task_info: Optional[Dict[str, Any]] = None,
        gender_map: Optional[Dict[str, int]] = None,
        race_map: Optional[Dict[str, int]] = None,
        ontology: Optional[Ontology] = None,
    ):
        self._tokenizer = tokenizer
        self._value_tokenizer = value_tokenizer
        self._att_tokenizer = att_tokenizer
        self._token_to_sub_time_token_mapping = token_to_sub_time_token_mapping
        self._concept_code_stats = concept_code_stats
        self._numeric_lab_stats = numeric_lab_stats
        self._numeric_event_statistics = NumericEventStatistics(numeric_lab_stats)
        self._categorical_lab_stats = categorical_lab_stats
        self._concept_name_mapping = concept_name_mapping
        self._oov_token_id = self._tokenizer.token_to_id(OUT_OF_VOCABULARY_TOKEN)
        self._padding_token_id = self._tokenizer.token_to_id(PAD_TOKEN)
        self._start_token_id = self._tokenizer.token_to_id(START_TOKEN)
        self._end_token_id = self._tokenizer.token_to_id(END_TOKEN)

        # Backward compatible with the old tokenizer
        self._linear_token_id = None
        if LINEAR_PROB_TOKEN in self._tokenizer.get_vocab():
            self._linear_token_id = self._tokenizer.token_to_id(LINEAR_PROB_TOKEN)
        else:
            self._linear_token_id = self._oov_token_id

        self._numeric_concept_ids = (
            self._numeric_event_statistics.get_numeric_concept_ids()
        )
        self._categorical_concept_ids = list(
            {t[0] for t in self._categorical_lab_stats.keys()}
        )
        self._padding_value_token_id = self._value_tokenizer.token_to_id(PAD_TOKEN)
        self._pretrained_concept_embedding_model = (
            pretrained_concept_embedding_model
            if pretrained_concept_embedding_model
            else PretrainedEmbeddings(None)
        )
        self._pretrained_concept_ids = [
            _
            for _ in self.get_vocab().keys()
            if self._pretrained_concept_embedding_model.is_concept_available(_)
        ]
        self._motor_task_info: Dict[str, Any] = (
            motor_task_info if motor_task_info is not None else {}
        )
        self._motor_time_to_event_codes = self._motor_task_info.get(
            "motor_time_to_event_codes", []
        )
        self._motor_code_to_id_mapping = {
            code: i for i, code in enumerate(sorted(self._motor_time_to_event_codes))
        }
        self._gender_map = gender_map if gender_map else {}
        self._race_map = race_map if race_map else {}
        self._ontology = ontology
        super().__init__()

    @property
    def pretrained_concept_ids(self):
        return self._pretrained_concept_ids

    @property
    def pretrained_token_ids(self):
        return self.encode(self._pretrained_concept_ids)

    @property
    def pretrained_embeddings(self):
        return np.asarray(
            [
                self._pretrained_concept_embedding_model.get_concept_embeddings(_)
                for _ in self._pretrained_concept_ids
            ]
        )

    @property
    def motor_tte_vocab_size(self) -> int:
        return len(self._motor_code_to_id_mapping)

    @property
    def vocab_size(self) -> int:
        return self._tokenizer.get_vocab_size()

    @property
    def value_vocab_size(self) -> int:
        return self._value_tokenizer.get_vocab_size()

    @property
    def time_token_vocab_size(self) -> int:
        return self._att_tokenizer.get_vocab_size()

    @property
    def gender_size(self) -> int:
        # Plus one for the unknown
        return len(self._gender_map) + 1

    @property
    def race_size(self) -> int:
        # Plus one for the unknown
        return len(self._race_map) + 1

    @property
    def pad_value_token_id(self):
        return self._padding_value_token_id

    @property
    def start_token_id(self):
        return self._start_token_id

    @property
    def end_token_id(self):
        return self._end_token_id

    @property
    def end_token(self):
        return END_TOKEN

    @property
    def eos_token(self):
        return END_TOKEN

    @property
    def eos_token_id(self):
        return self._end_token_id

    @property
    def pad_token_id(self):
        return self._padding_token_id

    @property
    def oov_token_id(self):
        return self._oov_token_id

    @property
    def pad_token(self):
        return PAD_TOKEN

    @property
    def linear_token_id(self) -> Optional[int]:
        return self._linear_token_id

    @property
    def linear_token(self) -> Optional[str]:
        if LINEAR_PROB_TOKEN in self._tokenizer.get_vocab():
            return LINEAR_PROB_TOKEN
        return None

    @property
    def vs_token_id(self):
        # We used VS for the historical data, currently, we use the new [VS] for the newer data
        # so we need to check both cases.
        if "VS" in self._tokenizer.get_vocab():
            return self._convert_token_to_id("VS")
        elif "[VS]" in self._tokenizer.get_vocab():
            return self._convert_token_to_id("[VS]")
        else:
            raise RuntimeError("The tokenizer does not contain either VS or [VS]")

    @property
    def ve_token_id(self):
        # We used VE for the historical data, currently, we use the new [VE] for the newer data
        # so we need to check both cases.
        if "VE" in self._tokenizer.get_vocab():
            return self._convert_token_to_id("VE")
        elif "[VE]" in self._tokenizer.get_vocab():
            return self._convert_token_to_id("[VE]")
        else:
            raise RuntimeError("The tokenizer does not contain either VE or [VE]")

    @property
    def numeric_concept_ids(self):
        return self._numeric_concept_ids

    @property
    def categorical_concept_ids(self):
        return self._categorical_concept_ids

    @property
    def lab_token_ids(self):
        reserved_tokens = [
            START_TOKEN,
            PAD_TOKEN,
            END_TOKEN,
            OUT_OF_VOCABULARY_TOKEN,
            LINEAR_PROB_TOKEN,
        ]
        lab_token_ids = self.encode(
            [
                concept_id
                for concept_id in self._numeric_concept_ids
                + self._categorical_concept_ids
                if concept_id not in reserved_tokens
            ]
        )
        return list(
            filter(lambda token_id: token_id != self._oov_token_id, lab_token_ids)
        )

    @property
    def token_to_time_token_mapping(self) -> Dict[int, List[int]]:
        default_mapping = {-1: [0, 0, 0]}
        default_mapping.update(
            {
                self._tokenizer.token_to_id(time_token): list(
                    map(self._att_tokenizer.token_to_id, sub_time_tokens)
                )
                for time_token, sub_time_tokens in self._token_to_sub_time_token_mapping.items()
            }
        )
        return default_mapping

    @property
    def pretrained_concept_embedding_model(self):
        return self._pretrained_concept_embedding_model

    def get_motor_time_bins(self, motor_num_time_pieces: int) -> List[int]:
        time_bins = np.percentile(
            self._motor_task_info["motor_event_times"].samples,
            np.linspace(0, 100, motor_num_time_pieces + 1),
        )
        time_bins[0] = 0
        time_bins[-1] = float("inf")
        return list(time_bins)

    def get_motor_token_id(self, concept_id: str) -> int:
        if not self.is_motor_time_to_event_code(concept_id):
            raise RuntimeError(f"Invalid motor concept id: {concept_id}")
        return self._motor_code_to_id_mapping[concept_id]

    def is_motor_time_to_event_code(self, future_concept_id: str) -> bool:
        if (
            self._motor_time_to_event_codes
            and future_concept_id in self._motor_time_to_event_codes
        ):
            return True
        return False

    def get_motor_parents(self, concept_id: str) -> List[str]:
        motor_codes = []
        if self._ontology is None:
            if self.is_motor_time_to_event_code(concept_id):
                motor_codes.append(concept_id)
        else:
            motor_codes.extend(
                [
                    p
                    for p in self._ontology.get_all_parents(concept_id)
                    if self.is_motor_time_to_event_code(p)
                ]
            )
        return motor_codes

    def get_vocab(self) -> Dict[str, int]:
        return self._tokenizer.get_vocab()

    def get_value_vocab(self) -> Dict[str, int]:
        return self._value_tokenizer.get_vocab()

    def encode(self, concept_ids, **kwargs) -> Sequence[int]:
        encoded = self._tokenizer.encode(concept_ids, is_pretokenized=True)
        return encoded.ids

    def decode(
        self, concept_token_ids: List[int], skip_special_tokens: bool = True, **kwargs
    ) -> List[str]:
        return self._tokenizer.decode(
            concept_token_ids, skip_special_tokens=skip_special_tokens
        ).split(" ")

    def encode_value(self, concept_values: Sequence[str]) -> Sequence[int]:
        encoded = self._value_tokenizer.encode(concept_values, is_pretokenized=True)
        return encoded.ids

    def decode_value(
        self, concept_value_token_ids: List[int], skip_special_tokens: bool = True
    ) -> List[str]:
        return self._value_tokenizer.decode(
            concept_value_token_ids, skip_special_tokens=skip_special_tokens
        ).split(" ")

    def encode_gender(self, gender: str) -> int:
        return self._gender_map.get(gender, 0)

    def encode_race(self, race: str) -> int:
        return self._race_map.get(race, 0)

    def add_token(self, tokens: Union[str, List[str]]) -> None:
        if isinstance(tokens, str):
            tokens = [tokens]
        vocab = self.get_vocab()
        self._tokenizer.add_tokens(
            [
                AddedToken(token, single_word=True, normalized=False)
                for token in tokens
                if token not in vocab
            ]
        )

    def _convert_token_to_id(self, token):
        """Converts a token (str) in an id using the vocab."""
        token_id = self._tokenizer.token_to_id(token)
        return token_id if token_id else self._oov_token_id

    def _convert_id_to_token(self, index):
        """Converts an index (integer) in a token (str) using the vocab."""
        token = self._tokenizer.id_to_token(index)
        return token if token else OUT_OF_VOCABULARY_TOKEN

    def convert_tokens_to_string(self, tokens):
        """Converts a sequence of tokens (string) in a single string."""
        out_string = " ".join([self._concept_name_mapping[t] for t in tokens])
        return out_string

    def save_pretrained(
        self,
        save_directory: Union[str, os.PathLike],
        push_to_hub: bool = False,
        **kwargs,
    ):
        """
        Save the Cehrbert tokenizer.

        This method make sure the batch processor can then be re-loaded using the
        .from_pretrained class method.

        Args:
            save_directory (`str` or `os.PathLike`): The path to a directory where the tokenizer will be saved.
            push_to_hub (`bool`, *optional*, defaults to `False`):
                Whether or not to push your model to the Hugging Face model hub after saving it. You can specify the
                repository you want to push to with `repo_id` (will default to the name of `save_directory` in your
                namespace).
            kwargs (`Dict[str, Any]`, *optional*):
                Additional key word arguments passed along to the [`PushToHubMixin.push_to_hub`] method.
        """
        assert not os.path.isfile(
            save_directory
        ), f"Provided path ({save_directory}) should be a directory, not a file"

        os.makedirs(save_directory, exist_ok=True)

        if push_to_hub:
            commit_message = kwargs.pop("commit_message", None)
            repo_id = kwargs.pop("repo_id", str(save_directory).split(os.path.sep)[-1])
            repo_id = self._create_repo(repo_id, **kwargs)
            files_timestamps = self._get_files_timestamps(save_directory)

        self._tokenizer.save(os.path.join(save_directory, TOKENIZER_FILE_NAME))

        self._value_tokenizer.save(
            os.path.join(save_directory, VALUE_TOKENIZER_FILE_NAME)
        )

        self._att_tokenizer.save(os.path.join(save_directory, TIME_TOKENIZER_FILE_NAME))

        with open(
            os.path.join(save_directory, TOKEN_TO_SUB_TIME_TOKEN_MAPPING_FILE_NAME), "w"
        ) as f:
            json.dump(self._token_to_sub_time_token_mapping, f)

        with open(os.path.join(save_directory, CONCEPT_STATS_FILE_NAME), "w") as f:
            json.dump(self._concept_code_stats, f)

        with open(os.path.join(save_directory, LAB_STATS_FILE_NAME), "wb") as f:
            lab_stats = {
                "numeric_lab_stats": self._numeric_lab_stats,
                "categorical_lab_stats": self._categorical_lab_stats,
            }
            pickle.dump(lab_stats, f)

        with open(
            os.path.join(save_directory, DEMOGRAPHICS_STATS_FILE_NAME), "wb"
        ) as f:
            pickle.dump(
                {
                    "gender_map": self._gender_map,
                    "race_map": self._race_map,
                },
                f,
            )

        with open(os.path.join(save_directory, CONCEPT_MAPPING_FILE_NAME), "w") as f:
            json.dump(self._concept_name_mapping, f)

        with open(
            os.path.join(save_directory, MOTOR_TIME_TO_EVENT_TASK_INFO_FILE_NAME), "wb"
        ) as f:
            pickle.dump(self._motor_task_info, f)

        if self._ontology is not None:
            with open(os.path.join(save_directory, ONTOLOGY_FILE_NAME), "wb") as f:
                pickle.dump(self._ontology, f)

        self._pretrained_concept_embedding_model.save(save_directory)

        if push_to_hub:
            self._upload_modified_files(
                save_directory,
                repo_id,
                files_timestamps,
                commit_message=commit_message,
                token=kwargs.get("token"),
            )

    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_name_or_path: Union[str, os.PathLike],
        **kwargs,
    ):
        """
        Load the CehrBert tokenizer.

        Parameters:
            pretrained_model_name_or_path (`str` or `os.PathLike`, *optional*):
                Can be either:
                    - A string, the *model id* of a pretrained model hosted inside a model repo on huggingface.co.
                      Valid model ids can be located at the root-level, like `bert-base-uncased`, or namespaced under a
                      user or organization name, like `dbmdz/bert-base-german-cased`.
                    - A path to a *directory* containing tokenization data saved using
                      [`save_pretrained`], e.g., `./my_data_directory/`.
            kwargs: Arguments for loading to pass to transformers.utils.hub.cached_file

        Returns:
            A CehrBert Tokenizer
        """

        is_legacy_tokenizer = CehrGptTokenizer.is_legacy_tokenizer(
            pretrained_model_name_or_path, **kwargs
        )

        # Load the concept tokenizer
        tokenizer_file = transformers.utils.hub.cached_file(
            pretrained_model_name_or_path, TOKENIZER_FILE_NAME, **kwargs
        )
        if not tokenizer_file:
            return None
        tokenizer = Tokenizer.from_file(tokenizer_file)

        # Load the concept_value_tokenizer
        if is_legacy_tokenizer:
            value_tokenizer = Tokenizer(
                WordLevel(unk_token=OUT_OF_VOCABULARY_TOKEN, vocab=dict())
            )
        else:
            value_tokenizer_file = transformers.utils.hub.cached_file(
                pretrained_model_name_or_path, VALUE_TOKENIZER_FILE_NAME, **kwargs
            )
            if not value_tokenizer_file:
                return None
            value_tokenizer = Tokenizer.from_file(value_tokenizer_file)

        # Load the ttt tokenizer
        att_tokenizer_file = transformers.utils.hub.cached_file(
            pretrained_model_name_or_path, TIME_TOKENIZER_FILE_NAME, **kwargs
        )
        if not att_tokenizer_file:
            return None
        att_tokenizer = Tokenizer.from_file(att_tokenizer_file)

        # Load the sub time token json file
        token_to_sub_time_token_mapping_file = transformers.utils.hub.cached_file(
            pretrained_model_name_or_path,
            TOKEN_TO_SUB_TIME_TOKEN_MAPPING_FILE_NAME,
            **kwargs,
        )
        if not token_to_sub_time_token_mapping_file:
            return None
        token_to_sub_time_token_mapping = load_json_file(
            token_to_sub_time_token_mapping_file
        )

        # Load the lab stats pickle file
        if is_legacy_tokenizer:
            legacy_lab_stats_file = transformers.utils.hub.cached_file(
                pretrained_model_name_or_path, LEGACY_LAB_STATS_FILE_NAME, **kwargs
            )
            if not legacy_lab_stats_file:
                return None
            # Support the old version of the numeric lab stats file
            lab_stats = {
                "numeric_lab_stats": load_json_file(legacy_lab_stats_file),
                "categorical_lab_stats": dict(),
            }
        else:
            lab_stats_file = transformers.utils.hub.cached_file(
                pretrained_model_name_or_path, LAB_STATS_FILE_NAME, **kwargs
            )
            if not lab_stats_file:
                return None

            with open(lab_stats_file, "rb") as file:
                lab_stats = pickle.load(file)
        try:
            # Load the demographics stats json file
            demographics_stats_file = transformers.utils.hub.cached_file(
                pretrained_model_name_or_path, DEMOGRAPHICS_STATS_FILE_NAME, **kwargs
            )
            if demographics_stats_file:
                with open(demographics_stats_file, "rb") as file:
                    demographics_stats = pickle.load(file)
            else:
                demographics_stats = None
        except EnvironmentError:
            LOG.warning(
                f"The %s files does not exist in %s, setting demographics_stats to None",
                DEMOGRAPHICS_STATS_FILE_NAME,
                pretrained_model_name_or_path,
            )
            demographics_stats = None

        # Load the concept_name json file
        concept_name_mapping_file = transformers.utils.hub.cached_file(
            pretrained_model_name_or_path, CONCEPT_MAPPING_FILE_NAME, **kwargs
        )
        if not concept_name_mapping_file:
            return None
        concept_name_mapping = load_json_file(concept_name_mapping_file)

        # Load the concept_code_stats json file
        concept_code_stats_mapping_file = transformers.utils.hub.cached_file(
            pretrained_model_name_or_path, CONCEPT_STATS_FILE_NAME, **kwargs
        )
        if not concept_code_stats_mapping_file:
            return None
        concept_code_stats = load_json_file(concept_code_stats_mapping_file)

        try:
            # Load the MOTOR time to event codes file
            motor_tte_task_info_file = transformers.utils.hub.cached_file(
                pretrained_model_name_or_path,
                MOTOR_TIME_TO_EVENT_TASK_INFO_FILE_NAME,
                **kwargs,
            )
            if motor_tte_task_info_file:
                with open(motor_tte_task_info_file, "rb") as file:
                    motor_task_info = pickle.load(file)
            else:
                motor_task_info = None
        except EnvironmentError:
            LOG.warning(
                f"The %s files does not exist in %s, setting motor_task_info to None",
                MOTOR_TIME_TO_EVENT_TASK_INFO_FILE_NAME,
                pretrained_model_name_or_path,
            )
            motor_task_info = None

        ontology = None
        try:
            ontology_file = transformers.utils.hub.cached_file(
                pretrained_model_name_or_path,
                ONTOLOGY_FILE_NAME,
                **kwargs,
            )
            if ontology_file:
                with open(ontology_file, "rb") as file:
                    ontology = pickle.load(file)
        except EnvironmentError | OSError:
            LOG.warning(
                "The ontology file %s does not existing in %s",
                ONTOLOGY_FILE_NAME,
                pretrained_model_name_or_path,
            )

        pretrained_embedding_model = PretrainedEmbeddings(pretrained_model_name_or_path)

        return CehrGptTokenizer(
            tokenizer,
            value_tokenizer,
            att_tokenizer,
            token_to_sub_time_token_mapping,
            concept_code_stats,
            lab_stats["numeric_lab_stats"],
            lab_stats["categorical_lab_stats"],
            concept_name_mapping,
            pretrained_embedding_model,
            motor_task_info,
            demographics_stats["gender_map"] if demographics_stats else None,
            demographics_stats["race_map"] if demographics_stats else None,
            ontology=ontology,
        )

    @classmethod
    def is_legacy_tokenizer(
        cls, pretrained_model_name_or_path: Union[str, os.PathLike], **kwargs
    ):
        try:
            legacy_lab_stats_file = transformers.utils.hub.cached_file(
                pretrained_model_name_or_path, LEGACY_LAB_STATS_FILE_NAME, **kwargs
            )
            return legacy_lab_stats_file is not None
        except Exception:
            return False

    @classmethod
    def expand_trained_tokenizer(
        cls,
        cehrgpt_tokenizer,
        dataset: Union[Dataset, DatasetDict],
        concept_name_mapping: Dict[str, str],
        data_args: DataTrainingArguments,
        pretrained_concept_embedding_model: PretrainedEmbeddings = None,
        num_motor_tasks: Optional[int] = None,
        apply_entropy_filter: bool = False,
        min_prevalence: float = 1 / 1000,
    ):
        if not isinstance(cehrgpt_tokenizer, CehrGptTokenizer):
            raise ValueError(
                "The existing cehrgpt must be an instance of CehrGptTokenizer"
            )

        cehrgpt_tokenizer_copy = copy.deepcopy(cehrgpt_tokenizer)

        new_tokenizer = CehrGptTokenizer.train_tokenizer(
            dataset=dataset,
            concept_name_mapping=concept_name_mapping,
            data_args=data_args,
            num_motor_tasks=num_motor_tasks,
            apply_entropy_filter=apply_entropy_filter,
            min_prevalence=min_prevalence,
        )

        new_tokens = set(new_tokenizer.get_vocab().keys()) - set(
            cehrgpt_tokenizer_copy.get_vocab().keys()
        )
        new_value_tokens = set(new_tokenizer.get_value_vocab().keys()) - set(
            cehrgpt_tokenizer_copy.get_value_vocab().keys()
        )
        new_att_tokens = set(new_tokenizer._att_tokenizer.get_vocab().keys()) - set(
            cehrgpt_tokenizer_copy._att_tokenizer.get_vocab().keys()
        )
        new_token_to_sub_time_token_mapping = (
            new_tokenizer._token_to_sub_time_token_mapping
        )
        new_numeric_lab_stats = new_tokenizer._numeric_lab_stats
        new_categorical_lab_stats = new_tokenizer._categorical_lab_stats
        new_concept_name_mapping = new_tokenizer._concept_name_mapping
        new_motor_time_to_event_codes = new_tokenizer._motor_time_to_event_codes

        # Add new tokens to the existing tokenizer
        cehrgpt_tokenizer_copy._tokenizer.add_tokens(
            [
                AddedToken(token, single_word=True, normalized=False)
                for token in new_tokens
            ]
        )
        # Add new tokens to the existing value tokenizer
        cehrgpt_tokenizer_copy._value_tokenizer.add_tokens(
            [
                AddedToken(token, single_word=True, normalized=False)
                for token in new_value_tokens
            ]
        )
        # Add new time tokens to the existing att tokenizer
        cehrgpt_tokenizer_copy._att_tokenizer.add_tokens(
            [
                AddedToken(token, single_word=True, normalized=False)
                for token in new_att_tokens
            ]
        )
        # Merge the time_token -> List[sub_time_tokens] mapping
        for time_token, sub_time_tokens in new_token_to_sub_time_token_mapping.items():
            if (
                time_token
                not in cehrgpt_tokenizer_copy._token_to_sub_time_token_mapping
            ):
                cehrgpt_tokenizer_copy._token_to_sub_time_token_mapping[time_token] = (
                    sub_time_tokens
                )

        # Merge numeric lab_stats
        cehrgpt_tokenizer_copy._numeric_lab_stats = cls.merge_numeric_lab_stats(
            cehrgpt_tokenizer_copy._numeric_lab_stats,
            new_numeric_lab_stats,
        )
        # Merge categorical lab_stats
        cehrgpt_tokenizer_copy._categorical_lab_stats = cls.merge_categorical_lab_stats(
            cehrgpt_tokenizer_copy._categorical_lab_stats,
            new_categorical_lab_stats,
        )

        # Merge concept_name_mapping
        for token, concept_name in new_concept_name_mapping.items():
            if token not in cehrgpt_tokenizer_copy._concept_name_mapping:
                cehrgpt_tokenizer_copy._concept_name_mapping[token] = concept_name

        # Merge motor_time_to_event_codes
        if (
            new_motor_time_to_event_codes
            and cehrgpt_tokenizer_copy._motor_time_to_event_codes
        ):
            for motor_time_to_event_code in new_motor_time_to_event_codes:
                if (
                    motor_time_to_event_code
                    not in cehrgpt_tokenizer_copy._motor_time_to_event_codes
                ):
                    cehrgpt_tokenizer_copy._motor_time_to_event_codes.append(
                        motor_time_to_event_code
                    )

        for gender in new_tokenizer._gender_map.keys():
            if gender not in cehrgpt_tokenizer_copy._gender_map:
                cehrgpt_tokenizer_copy._gender_map[gender] = len(
                    cehrgpt_tokenizer_copy._gender_map
                )

        for race in new_tokenizer._race_map.keys():
            if race not in cehrgpt_tokenizer_copy._race_map:
                cehrgpt_tokenizer_copy._race_map[race] = len(
                    cehrgpt_tokenizer_copy._race_map
                )

        return CehrGptTokenizer(
            tokenizer=cehrgpt_tokenizer_copy._tokenizer,
            value_tokenizer=cehrgpt_tokenizer_copy._value_tokenizer,
            att_tokenizer=cehrgpt_tokenizer_copy._att_tokenizer,
            token_to_sub_time_token_mapping=cehrgpt_tokenizer_copy._token_to_sub_time_token_mapping,
            concept_code_stats=cehrgpt_tokenizer_copy._concept_code_stats,
            numeric_lab_stats=cehrgpt_tokenizer_copy._numeric_lab_stats,
            categorical_lab_stats=cehrgpt_tokenizer_copy._categorical_lab_stats,
            concept_name_mapping=cehrgpt_tokenizer_copy._concept_name_mapping,
            pretrained_concept_embedding_model=pretrained_concept_embedding_model,
            motor_task_info=cehrgpt_tokenizer_copy._motor_task_info,
            gender_map=cehrgpt_tokenizer_copy._gender_map,
            race_map=cehrgpt_tokenizer_copy._race_map,
            ontology=cehrgpt_tokenizer_copy._ontology,
        )

    @classmethod
    def merge_numeric_lab_stats(
        cls,
        lab_stats_existing: List[Dict[str, Any]],
        lab_stats_new: List[Dict[str, Any]],
    ) -> List[Dict[str, Any]]:

        lab_stats_existing_mapping = {
            (lab_stat["concept_id"], lab_stat["unit"]): lab_stat
            for lab_stat in lab_stats_existing
        }
        for lab_stat in lab_stats_new:
            concept_unit_pair = (lab_stat["concept_id"], lab_stat["unit"])
            if concept_unit_pair in lab_stats_existing_mapping:
                existing = OnlineStatistics()
                existing.count = lab_stats_existing_mapping[concept_unit_pair]["count"]
                existing.current_mean = lab_stats_existing_mapping[concept_unit_pair][
                    "mean"
                ]
                existing.variance = (
                    lab_stats_existing_mapping[concept_unit_pair]["std"] ** 2
                    * existing.count
                )
                new = OnlineStatistics()
                new.count = lab_stat["count"]
                new.current_mean = lab_stat["mean"]
                new.variance = lab_stat["std"] ** 2 * new.count
                existing.combine(new)
                lab_stats_existing_mapping[concept_unit_pair]["mean"] = existing.mean()
                lab_stats_existing_mapping[concept_unit_pair][
                    "std"
                ] = existing.standard_deviation()
                lab_stats_existing_mapping[concept_unit_pair]["count"] = existing.count
                # recreate the bins
                sample = create_sample_from_bins(
                    lab_stats_existing_mapping[concept_unit_pair]["bins"]
                )
                sample.extend(create_sample_from_bins(lab_stat["bins"]))
                lab_stats_existing_mapping[concept_unit_pair]["bins"] = (
                    create_bins_with_spline(sample, NUM_OF_BINS, DEGREE_OF_FREEDOM)
                )

            else:
                if lab_stat["count"] > 0:
                    lab_stats_existing_mapping[concept_unit_pair] = lab_stat

        return list(lab_stats_existing_mapping.values())

    @classmethod
    def merge_categorical_lab_stats(
        cls,
        categorical_lab_stats_existing: Dict[Tuple[str, str], int],
        categorical_lab_stats_new: Dict[Tuple[str, str], int],
    ) -> Dict[Tuple[str, str], int]:
        for (concept_id, concept_as_value), count in categorical_lab_stats_new.items():
            if (concept_id, concept_as_value) not in categorical_lab_stats_new:
                categorical_lab_stats_existing[(concept_id, concept_as_value)] = 0
            categorical_lab_stats_existing[(concept_id, concept_as_value)] += count
        return categorical_lab_stats_existing

    @classmethod
    def train_tokenizer(
        cls,
        dataset: Union[Dataset, DatasetDict],
        concept_name_mapping: Dict[str, str],
        data_args: DataTrainingArguments,
        pretrained_concept_embedding_model: PretrainedEmbeddings = None,
        num_motor_tasks: Optional[int] = None,
        apply_entropy_filter: bool = False,
        min_prevalence: float = 1 / 1000,
        ontology: Optional[Ontology] = None,
    ):
        """
        Train a huggingface word level tokenizer.

        To use their tokenizer, we need to concatenate all the concepts
        together and treat it as a sequence.
        """

        if isinstance(dataset, DatasetDict):
            dataset = dataset["train"]

        LOG.info("Calculating data statistics")
        cehrgpt_data_statistics = compute_statistics(dataset, data_args, ontology)
        numeric_lab_stats = cehrgpt_data_statistics["numeric_lab_stats"]
        categorical_lab_stats = cehrgpt_data_statistics["categorical_lab_stats"]
        original_concept_codes = cehrgpt_data_statistics["original_concept_codes"]
        all_concept_code_stats = cehrgpt_data_statistics["all_concept_code_stats"]
        all_concept_code_entropies = cehrgpt_data_statistics[
            "all_concept_code_entropies"
        ]

        if apply_entropy_filter:
            min_prevalence = max(1e-8, min_prevalence)
            min_entropy = (
                np.log(1 - min_prevalence) * (1 - min_prevalence)
                + np.log(min_prevalence) * min_prevalence
            )
            qualified_codes = [
                k
                for k in original_concept_codes
                if all_concept_code_entropies[k] <= min_entropy
                or not is_clinical_event(k, data_args.is_data_in_meds)
            ]
        else:
            qualified_codes = [
                k
                for k in original_concept_codes
                if min_prevalence <= all_concept_code_stats[k]
                or not is_clinical_event(k, data_args.is_data_in_meds)
            ]

        # Create the tokenizer now
        LOG.info("Training the tokenizer for concepts")
        concept_tokenizer = cls.train_concept_tokenizer(
            dataset,
            feature_name="concept_ids",
            special_tokens=[
                PAD_TOKEN,
                OUT_OF_VOCABULARY_TOKEN,
                START_TOKEN,
                END_TOKEN,
                LINEAR_PROB_TOKEN,
            ],
            unk_token=OUT_OF_VOCABULARY_TOKEN,
            data_args=data_args,
            qualified_codes=qualified_codes,
        )
        concept_value_column = "concept_as_values"
        for row in dataset:
            if concept_value_column not in row:
                concept_value_column = "concept_values"
            break
        LOG.info("Training the tokenizer for values")
        value_tokenizer = cls.train_concept_tokenizer(
            dataset,
            feature_name=concept_value_column,
            special_tokens=[OUT_OF_VOCABULARY_TOKEN, PAD_TOKEN],
            unk_token=OUT_OF_VOCABULARY_TOKEN,
            data_args=data_args,
        )
        value_tokenizer.add_tokens(
            [
                AddedToken(_, single_word=True, normalized=False)
                for _ in [create_value_bin(_) for _ in range(NUM_OF_BINS)]
                + [UNKNOWN_BIN, NONE_BIN]
            ]
        )

        # We will train a tokenizer specifically for time intervals
        sub_time_token_data = []
        token_to_sub_time_token_mapping = collections.defaultdict(list)
        vocab = concept_tokenizer.get_vocab()
        for token, token_id in vocab.items():
            if is_att_token(token):
                time_interval = extract_time_interval_in_days(token)
                time_tuple = convert_time_interval_to_time_tuple(
                    time_interval, is_inpatient_att_token(token)
                )
                token_to_sub_time_token_mapping[token] = list(time_tuple)
                sub_time_token_data.append(" ".join(time_tuple))

        att_tokenizer = Tokenizer(
            WordLevel(unk_token=OUT_OF_VOCABULARY_TOKEN, vocab=dict())
        )
        att_tokenizer.pre_tokenizer = WhitespaceSplit()
        att_trainer = WordLevelTrainer(
            special_tokens=[OUT_OF_VOCABULARY_TOKEN],
            vocab_size=data_args.vocab_size,
            min_frequency=0,
            show_progress=True,
        )
        att_tokenizer.train_from_iterator(sub_time_token_data, trainer=att_trainer)

        # Prune concept_name_mapping
        concept_name_mapping = {
            concept_id: concept_name_mapping[concept_id]
            for concept_id in vocab.keys()
            if concept_id in concept_name_mapping
        }

        motor_task_info = None
        if num_motor_tasks:
            LOG.info("Computing the MOTOR TTE statistics")
            allowed_motor_codes = get_allowed_motor_codes(
                original_concept_codes, ontology
            )
            motor_tte_statistics = compute_motor_tte_statistics(
                dataset, data_args, allowed_motor_codes, ontology
            )
            motor_time_to_event_codes = []
            for concept_id, _ in sorted(
                all_concept_code_entropies.items(), key=lambda t: t[1]
            ):
                if concept_id not in allowed_motor_codes:
                    continue
                tte_stats = motor_tte_statistics["task_tte_stats"][concept_id]
                censor_stats = motor_tte_statistics["task_censor_stats"][concept_id]
                frac_events = tte_stats / (tte_stats + censor_stats)

                if frac_events < 1 / 1000:
                    LOG.info(
                        "Ran into very rare task %s with %s", concept_id, frac_events
                    )
                    continue

                if len(motor_time_to_event_codes) < num_motor_tasks:
                    motor_time_to_event_codes.append(concept_id)
                else:
                    break

            LOG.info(
                f"{len(motor_time_to_event_codes)} number of tasks have been added as MOTOR tasks"
            )
            motor_task_info = {
                "motor_event_times": motor_tte_statistics["motor_event_times"],
                "task_tte_stats": motor_tte_statistics["task_tte_stats"],
                "task_censor_stats": motor_tte_statistics["task_censor_stats"],
                "motor_time_to_event_codes": motor_time_to_event_codes,
            }

        gender_map = {
            gender: i + 1
            for i, gender in enumerate(sorted(cehrgpt_data_statistics["gender_list"]))
        }
        race_map = {
            race: i + 1
            for i, race in enumerate(sorted(cehrgpt_data_statistics["race_list"]))
        }

        return CehrGptTokenizer(
            concept_tokenizer,
            value_tokenizer,
            att_tokenizer,
            token_to_sub_time_token_mapping,
            all_concept_code_stats,
            numeric_lab_stats,
            categorical_lab_stats,
            concept_name_mapping,
            pretrained_concept_embedding_model,
            motor_task_info,
            gender_map,
            race_map,
            ontology,
        )

    @classmethod
    def train_concept_tokenizer(
        cls,
        dataset,
        feature_name,
        special_tokens: List[str],
        unk_token,
        data_args,
        qualified_codes: Optional[List[str]] = None,
    ):
        # Use the Fast Tokenizer from the Huggingface tokenizers Rust implementation.
        # https://github.com/huggingface/tokenizers
        concept_tokenizer = Tokenizer(WordLevel(unk_token=unk_token, vocab=dict()))
        concept_tokenizer.pre_tokenizer = WhitespaceSplit()
        concept_trainer = WordLevelTrainer(
            special_tokens=special_tokens,
            vocab_size=data_args.vocab_size,
            min_frequency=data_args.min_frequency,
            show_progress=True,
        )
        batch_concat_concepts_partial_func = partial(
            cls.batch_concat_concepts,
            feature_name=feature_name,
            qualified_codes=qualified_codes,
        )
        if data_args.streaming:
            concatenated_features = dataset.map(
                batch_concat_concepts_partial_func,
                batched=True,
                batch_size=data_args.preprocessing_batch_size,
            )

            def batched_generator():
                iterator = iter(concatenated_features)
                while True:
                    batch = list(islice(iterator, data_args.preprocessing_batch_size))
                    if not batch:
                        break
                    yield [example[feature_name] for example in batch]

            # We pass a generator of list of texts (concatenated concept_ids) to train_from_iterator
            # for efficient training
            generator = batched_generator()
        else:
            concatenated_features = dataset.map(
                batch_concat_concepts_partial_func,
                num_proc=data_args.preprocessing_num_workers,
                batched=True,
                batch_size=data_args.preprocessing_batch_size,
                remove_columns=dataset.column_names,
            )
            generator = concatenated_features[feature_name]
        concept_tokenizer.train_from_iterator(generator, trainer=concept_trainer)
        return concept_tokenizer

    def normalize(self, concept_id: str, unit: str, concept_value: float) -> str:
        return self._numeric_event_statistics.normalize(concept_id, unit, concept_value)

    def denormalize(self, concept_id: str, value_bin: str) -> Tuple[float, str]:
        return self._numeric_event_statistics.denormalize(concept_id, value_bin)

    @classmethod
    def batch_concat_concepts(
        cls,
        records: Dict[str, List],
        feature_name: str,
        qualified_codes: Optional[List[str]] = None,
    ) -> Dict[str, List]:
        def filter_token(t: str) -> bool:
            """
            If the token is None or not string, return False.

            When qualified_codes is provided, t must be in
            qualified_codes to be valid, otherwise the tokens are always valid

            :param t:
            :return:
            """
            if t is None or not isinstance(t, str):
                return False
            if qualified_codes:
                return t in qualified_codes
            return True

        return {
            feature_name: [
                " ".join(filter(filter_token, tokens))
                for tokens in records[feature_name]
            ]
        }



import math
from collections import Counter
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from cehrbert_data.decorators.patient_event_decorator_base import time_month_token
from transformers import GenerationConfig

from cehrgpt.gpt_utils import (
    extract_time_interval_in_days,
    is_att_token,
    is_visit_end,
    is_visit_start,
)
from cehrgpt.models.hf_cehrgpt import CEHRGPT2LMHeadModel
from cehrgpt.models.tokenization_hf_cehrgpt import CehrGptTokenizer


@dataclass
class TimeToEvent:
    average_time: float
    median_time: float
    standard_deviation: float
    most_likely_time: str
    num_of_simulations: int
    time_intervals: List[int]
    outcome_events: List[str]
    time_interval_probability_table: List[Dict[str, Any]]


def create_time_to_event(
    time_event_tuples: List[Tuple[str, int]], num_of_simulations: int
) -> TimeToEvent:
    outcome_events, time_intervals = zip(*time_event_tuples)
    time_buckets = [time_month_token(_) for _ in time_intervals]
    time_bucket_counter = Counter(time_buckets)
    most_common_item = time_bucket_counter.most_common(1)[0][0]
    total_count = sum(time_bucket_counter.values())
    # Generate the probability table
    probability_table = {
        item: count / total_count for item, count in time_bucket_counter.items()
    }
    sorted_probability_table = [
        {"time_interval": k, "probability": v}
        for k, v in sorted(probability_table.items(), key=lambda x: x[1], reverse=True)
    ]
    return TimeToEvent(
        time_intervals=time_intervals,
        outcome_events=outcome_events,
        average_time=np.mean(time_intervals),
        median_time=np.median(time_intervals),
        standard_deviation=np.std(time_intervals),
        most_likely_time=most_common_item,
        num_of_simulations=num_of_simulations,
        time_interval_probability_table=sorted_probability_table,
    )


class TimeToEventModel:
    def __init__(
        self,
        tokenizer: CehrGptTokenizer,
        model: CEHRGPT2LMHeadModel,
        outcome_events: List[str],
        generation_config: GenerationConfig,
        device: torch.device = torch.device("cpu"),
        batch_size: int = 32,
    ):
        self.tokenizer = tokenizer
        self.model = model.eval()
        self.generation_config = generation_config
        self.outcome_events = outcome_events
        self.device = device
        self.batch_size = batch_size
        self.max_sequence = model.config.n_positions

    def is_outcome_event(self, token: str):
        return token in self.outcome_events

    def simulate(
        self,
        partial_history: Union[np.ndarray, List[str]],
    ) -> List[List[str]]:
        token_ids = self.tokenizer.encode(partial_history)
        prompt = torch.tensor(token_ids).unsqueeze(0).to(self.device)

        simulated_sequences = []
        num_iters = max(
            math.ceil(self.generation_config.num_return_sequences / self.batch_size), 1
        )
        old_num_return_sequences = self.generation_config.num_return_sequences
        self.generation_config.num_return_sequences = min(
            self.batch_size, old_num_return_sequences
        )
        with torch.no_grad():
            for _ in range(num_iters):
                results = self.model.generate(
                    inputs=prompt,
                    generation_config=self.generation_config,
                )
                # Clear the cache
                torch.cuda.empty_cache()
                # Add the sequences to the result array
                simulated_sequences.extend(
                    [
                        self.tokenizer.decode(seq.cpu().numpy())
                        for seq in results.sequences
                    ]
                )

        self.generation_config.num_return_sequences = old_num_return_sequences
        return simulated_sequences

    def predict_time_to_events(
        self,
        partial_history: Union[np.ndarray, list],
        future_visit_start: int = 0,
        future_visit_end: int = -1,
        prediction_window_start: int = 0,
        prediction_window_end: int = 365,
        debug: bool = False,
        max_n_trial: int = 2,
    ) -> Optional[TimeToEvent]:
        patient_history_length = len(partial_history)
        time_event_tuples = []
        seqs_failed_to_convert = []
        n_trial = 0
        num_return_sequences = self.generation_config.num_return_sequences
        max_new_tokens = self.generation_config.max_new_tokens
        while (
            len(time_event_tuples) < self.generation_config.num_return_sequences
            and n_trial < max_n_trial
        ):
            self.generation_config.num_return_sequences = num_return_sequences - len(
                time_event_tuples
            )
            # self.generation_config.max_new_tokens = max_new_tokens * (n_trial + 1)
            simulated_seqs = self.simulate(partial_history)
            n_trial += 1
            for seq in simulated_seqs:
                visit_counter = 0
                time_delta = 0
                success = False
                for next_token in seq[patient_history_length:]:
                    visit_counter += int(is_visit_start(next_token))
                    if (
                        visit_counter > future_visit_end != -1
                        or time_delta > prediction_window_end != -1
                    ):
                        time_event_tuples.append(("0", time_delta))
                        success = True
                        break
                    if is_att_token(next_token):
                        time_delta += extract_time_interval_in_days(next_token)
                    elif (
                        visit_counter >= future_visit_start
                        and time_delta >= prediction_window_start
                    ) and self.is_outcome_event(next_token):
                        time_event_tuples.append((next_token, time_delta))
                        success = True
                        break
                if not success:
                    # This indicates the generated sequence did not satisfy the criteria
                    if future_visit_end != -1 or prediction_window_end != -1:
                        seqs_failed_to_convert.append(seq[patient_history_length:])
                    else:
                        time_event_tuples.append(("0", time_delta))

        self.generation_config.num_return_sequences = num_return_sequences
        self.generation_config.max_new_tokens = max_new_tokens

        if debug:
            print(f"seqs_failed_to_convert: {seqs_failed_to_convert}")

        # Count the occurrences of each time tokens for each concept
        return (
            create_time_to_event(time_event_tuples, len(time_event_tuples))
            if len(time_event_tuples) > 0
            else None
        )

    @staticmethod
    def get_generation_config(
        tokenizer: CehrGptTokenizer,
        max_length: int,
        num_return_sequences: int,
        top_p: float = 1.0,
        top_k: int = 300,
        temperature: float = 1.0,
        repetition_penalty: float = 1.0,
        epsilon_cutoff: float = 0.0,
        max_new_tokens: int = 128,
    ) -> GenerationConfig:
        return GenerationConfig(
            max_length=max_length,
            max_new_tokens=max_new_tokens,
            num_return_sequences=num_return_sequences,
            temperature=temperature,
            repetition_penalty=repetition_penalty,
            epsilon_cutoff=epsilon_cutoff,
            top_p=top_p,
            top_k=top_k,
            bos_token_id=tokenizer.end_token_id,
            eos_token_id=tokenizer.end_token_id,
            pad_token_id=tokenizer.pad_token_id,
            do_sample=True,
            use_cache=True,
            return_dict_in_generate=True,
            output_attentions=False,
            output_hidden_states=False,
            output_scores=False,
            renormalize_logits=True,
        )



import random
import re
from datetime import date, datetime, timedelta, timezone
from typing import List, Optional, Sequence, Tuple, Union

import numpy as np
from cehrbert_data.const.artificial_tokens import DEATH_TOKEN
from meds import death_code
from transformers.utils import logging

from cehrgpt.cehrgpt_args import SamplingStrategy
from cehrgpt.models.special_tokens import (
    DISCHARGE_CONCEPT_IDS,
    END_TOKEN,
    VISIT_CONCEPT_IDS,
)

# Regular expression pattern to match inpatient attendance tokens
MEDS_CODE_PATTERN = re.compile(r".*/.*")
INPATIENT_ATT_PATTERN = re.compile(r"(?:VS-|i-)D(\d+)(?:-VE)?")
DEMOGRAPHIC_PROMPT_SIZE = 4
logger = logging.get_logger("transformers")


class RandomSampleCache:
    def __init__(
        self,
        data_indices: Sequence[int],
        cache_size: int,
        sample_weights: Sequence[float] = None,
    ):
        """
        Initialize the RandomSampleCache.

        :param data_indices: Sequence of data indices to sample from.
        :param cache_size: Size of the cache.
        :param sample_weights: Optional sequence of weights for sampling.
        """
        self._data_indices = data_indices
        self._sample_weights = sample_weights
        self._cache_size = cache_size
        self._cache = []

        if self._sample_weights is not None:
            assert sum(self._sample_weights) - 1 < 1e-8

    def next(self):
        """
        Get the next sample from the cache.

        If the cache is empty, refill it.

        :return: A sampled data index.
        """
        if not self._cache:
            if self._sample_weights is not None:
                self._cache.extend(
                    random.choices(
                        self._data_indices,
                        k=self._cache_size,
                        weights=self._sample_weights,
                    )
                )
            else:
                self._cache.extend(
                    random.choices(self._data_indices, k=self._cache_size)
                )
        return self._cache.pop()


def construct_time_sequence(
    concept_ids: List[str], epoch_times: Optional[List[Union[int, float]]] = None
) -> List[float]:
    if epoch_times is not None:
        return epoch_times

    if concept_ids[0].lower().startswith("year"):
        year_str = concept_ids[0].split(":")[1]
    else:
        year_str = "1985"

    datetime_cursor = datetime(
        int(year_str), month=1, day=1, hour=0, minute=0, second=0
    ).replace(tzinfo=timezone.utc)
    epoch_times = []
    for concept_id in concept_ids:
        if is_att_token(concept_id):
            att_days = extract_time_interval_in_days(concept_id)
            datetime_cursor += timedelta(days=att_days)
        epoch_times.append(datetime_cursor.timestamp())
    return epoch_times


def construct_age_sequence(
    concept_ids: List[str], ages: Optional[List[int]] = None
) -> List[int]:
    if ages is not None:
        return ages
    elif concept_ids[1].lower().startswith("age"):
        age_str = concept_ids[1].split(":")[1]
        try:
            age = max(int(age_str), 0)
        except ValueError:
            logger.warning(
                "Age is not an integer: %s. Setting it to default value 0", age_str
            )
            age = 0
        ages = []
        time_delta = 0
        for concept_id in concept_ids:
            if is_att_token(concept_id):
                time_delta += extract_time_interval_in_days(concept_id)
            ages.append(age + time_delta // 365)
        return ages
    else:
        logger.warning(
            "The second token is not a valid age token. The first 4 tokens are: %s. "
            "Trying to fall back to ages, but it is not valid either %s. "
            "Fall back to a zero vector [0, 0, 0, ...., 0]",
            concept_ids[:4],
            ages,
        )
        return np.zeros_like(concept_ids, dtype=int).tolist()


def multiple_of_10(n: int) -> int:
    return ((n // 10) + 1) * 10


def encode_demographics(
    age: int, gender: int, race: int, max_age=200, max_gender=10, max_race=10
) -> int:
    assert 0 <= age < max_age, f"age: {age}"
    assert 0 <= gender < max_gender, f"gender: {gender}"
    assert 0 <= race < max_race, f"race: {race}"
    return age + max_age * gender + max_age * max_gender * race


def collect_demographic_prompts_at_visits(patient_history: List[str]):
    demographic_prompts_at_visits = []
    start_year, start_age, start_gender, start_race = patient_history[
        :DEMOGRAPHIC_PROMPT_SIZE
    ]
    try:
        start_year = int(start_year.split(":")[1])
        start_age = int(start_age.split(":")[1])
        valid_prompt = True
    except IndexError | ValueError:
        start_year = 1900
        start_age = 0
        valid_prompt = False
    data_cursor = date(int(start_year), 1, 1)
    birth_date = date(start_year - start_age, 1, 1)
    for i, current_token in enumerate(patient_history):
        if is_visit_start(current_token):
            reconstructed_year = (
                f"year:{data_cursor.year}" if valid_prompt else "year:unknown"
            )
            reconstructed_age = (
                f"age:{data_cursor.year - birth_date.year}"
                if valid_prompt
                else "age:unknown"
            )
            demographic_prompts_at_visits.append(
                (
                    i,
                    (
                        reconstructed_year,
                        reconstructed_age,
                        start_gender,
                        start_race,
                    ),
                )
            )
        elif is_att_token(current_token):
            att_date_delta = extract_time_interval_in_days(current_token)
            data_cursor = data_cursor + timedelta(days=att_date_delta)
    return demographic_prompts_at_visits


def random_slice_gpt_sequence(concept_ids, max_seq_len):
    """
    Randomly slice a GPT sequence.

    :param concept_ids: List of concept IDs.
    :param max_seq_len: Maximum sequence length.
    :return: Tuple containing start index, end index, and demographic tokens.
    """
    seq_length = len(concept_ids)
    starting_points = []
    start_year, start_age, start_gender, start_race = [
        _ for _ in concept_ids[:DEMOGRAPHIC_PROMPT_SIZE]
    ]
    try:
        start_year = int(start_year.split(":")[1])
        start_age = int(start_age.split(":")[1])
        data_cursor = date(int(start_year), 1, 1)
        birth_date = date(start_year - start_age, 1, 1)
        for i in range(
            DEMOGRAPHIC_PROMPT_SIZE,
            min(seq_length, seq_length - max_seq_len + DEMOGRAPHIC_PROMPT_SIZE),
        ):
            current_token = concept_ids[i]
            if is_visit_start(current_token):
                starting_points.append(
                    (i, data_cursor.year, data_cursor.year - birth_date.year)
                )
            elif is_att_token(current_token):
                att_date_delta = extract_time_interval_in_days(current_token)
                data_cursor = data_cursor + timedelta(days=att_date_delta)

        if len(starting_points) == 0:
            return 0, 0, concept_ids[:DEMOGRAPHIC_PROMPT_SIZE]

        random_starting_index, random_starting_year, random_starting_age = (
            random.choice(starting_points)
        )
        demographic_tokens = [
            f"year:{random_starting_year}",
            f"age:{random_starting_age}",
            start_gender,
            start_race,
        ]
        # Remove the number of demographic tokens
        random_end_index = random_starting_index
        for i in reversed(
            range(
                random_starting_index,
                random_starting_index + max_seq_len - DEMOGRAPHIC_PROMPT_SIZE,
            )
        ):
            current_token = concept_ids[i]
            if is_visit_end(current_token):
                random_end_index = i
                break
        return random_starting_index, random_end_index, demographic_tokens

    except Exception:
        return 0, max_seq_len - 1, []


def get_cehrgpt_output_folder(args, cehrgpt_tokenizer) -> str:
    if args.sampling_strategy == SamplingStrategy.TopKStrategy.value:
        folder_name = f"top_k{args.top_k}"
        args.top_p = 1.0
    elif args.sampling_strategy == SamplingStrategy.TopPStrategy.value:
        folder_name = f"top_p{int(args.top_p * 10000)}"
        args.top_k = cehrgpt_tokenizer.vocab_size
    elif args.sampling_strategy == SamplingStrategy.TopMixStrategy.value:
        folder_name = f"top_mix_p{int(args.top_p * 10000)}_k{args.top_k}"
    else:
        raise RuntimeError(
            "sampling_strategy has to be one of the following three options [TopKStrategy, TopPStrategy, TopMixStrategy]"
        )
    if args.temperature != 1.0:
        folder_name = f"{folder_name}_temp_{int(args.temperature * 10000)}"
    if args.repetition_penalty != 1.0:
        folder_name = (
            f"{folder_name}_repetition_penalty_{int(args.repetition_penalty * 10000)}"
        )
    if args.num_beams > 1:
        folder_name = f"{folder_name}_num_beams_{int(args.num_beams)}"
    if args.num_beam_groups > 1:
        folder_name = f"{folder_name}_num_beam_groups_{int(args.num_beam_groups)}"
    if args.epsilon_cutoff > 0.0:
        folder_name = (
            f"{folder_name}_epsilon_cutoff_{int(args.epsilon_cutoff * 100000)}"
        )
    return folder_name


def is_clinical_event(token: str, meds: bool = False) -> bool:
    if token.isnumeric():
        return True
    if token in [DEATH_TOKEN, death_code]:
        return True
    if meds:
        return bool(MEDS_CODE_PATTERN.match(token))
    return False


def is_visit_start(token: str):
    """
    Check if the token indicates the start of a visit.

    :param token: Token to check.
    :return: True if the token is a visit start token, False otherwise.
    """
    return token in ["VS", "[VS]"]


def is_visit_end(token: str) -> bool:
    return token in ["VE", "[VE]"]


def is_inpatient_hour_token(token: str) -> bool:
    return token.startswith("i-H")


def extract_time_interval_in_hours(token: str) -> int:
    try:
        hour = int(token[3:])
        return hour
    except ValueError:
        return 0


def is_att_token(token: str):
    """
    Check if the token is an attention token.

    :param token: Token to check.
    :return: True if the token is an attention token, False otherwise.
    """
    if bool(re.match(r"^D\d+", token)):  # day tokens
        return True
    elif bool(re.match(r"^W\d+", token)):  # week tokens
        return True
    elif bool(re.match(r"^M\d+", token)):  # month tokens
        return True
    elif bool(re.match(r"^Y\d+", token)):  # year tokens
        return True
    elif token == "LT":
        return True
    elif token[:3] == "VS-":  # VS-D7-VE
        return True
    elif token[:2] == "i-" and not token.startswith(
        "i-H"
    ):  # i-D7 and exclude hour tokens
        return True
    return False


def is_artificial_token(token: str) -> bool:
    if token in VISIT_CONCEPT_IDS:
        return True
    if token in DISCHARGE_CONCEPT_IDS:
        return True
    if is_visit_start(token):
        return True
    if is_visit_end(token):
        return True
    if is_att_token(token):
        return True
    if token == END_TOKEN:
        return True

    return False


def is_inpatient_att_token(token: str):
    """
    Check if the token is an inpatient ATT token.

    :param token: Token to check.
    :return: True if the token is an inpatient ATT token, False otherwise.
    """
    return INPATIENT_ATT_PATTERN.match(token)


def extract_time_interval_in_days(token: str):
    """
    Extract the time interval in days from a token.

    :param token: Token to extract from.
    :return: Time interval in days.
    :raises ValueError: If the token is invalid.
    """
    try:
        if token[0] == "D":  # day tokens
            return int(token[1:])
        elif token[0] == "W":  # week tokens
            return int(token[1:]) * 7
        elif token[0] == "M":  # month tokens
            return int(token[1:]) * 30
        elif token[0] == "Y":  # year tokens
            return int(token[1:]) * 365
        elif token == "LT":
            return 365 * 3
        elif token[:3] == "VS-":  # VS-D7-VE
            part = token.split("-")[1]
            if part.startswith("LT"):
                return 365 * 3
            return int(part[1:])
        elif token[:2] == "i-":  # i-D7
            part = token.split("-")[1]
            if part.startswith("LT"):
                return 365 * 3
            return int(token.split("-")[1][1:])
    except Exception:
        raise ValueError(f"Invalid time token: {token}")
    raise ValueError(f"Invalid time token: {token}")


def convert_time_interval_to_time_tuple(
    time_interval: int, is_inpatient: bool
) -> Tuple[str, str, str]:
    """
    Convert a time interval to a tuple of time tokens.

    :param time_interval: Time interval in days.
    :param is_inpatient: Whether the interval is for an inpatient.
    :return: Tuple of year, month, and day tokens.
    """
    assert time_interval >= 0, "the time interval must equal and greater than zero"
    year = time_interval // 365
    month = time_interval % 365 // 30
    day = time_interval % 365 % 30
    year_token = f"year:{year}"
    month_token = f"month:{month}"
    day_token = f"i-day:{day}" if is_inpatient else f"day:{day}"
    return year_token, month_token, day_token


def generate_artificial_time_tokens():
    """
    Generate all the time tokens used in training.

    :return: List of time tokens.
    """
    day_tokens = [f"D{i}" for i in range(2000)]
    week_tokens = [f"W{i}" for i in range(4)]
    month_tokens = [f"M{i}" for i in range(12)]
    long_term_tokens = ["LT"]
    return day_tokens + week_tokens + month_tokens + long_term_tokens



#!/usr/bin/env python
"""
Treatment Pathways Study Script.

Analyzes hypertension treatment pathways using OMOP CDM data
"""

import argparse
import os
import sys

from pyspark.sql import SparkSession
from pyspark.sql import functions as f


def parse_arguments():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Analyze hypertension treatment pathways using OMOP CDM data"
    )

    parser.add_argument(
        "--omop-folder",
        required=True,
        help="Path to OMOP CDM data folder containing parquet files",
    )

    parser.add_argument(
        "--output-folder", required=True, help="Output folder for results"
    )

    parser.add_argument(
        "--app-name",
        default="HTN_Treatment_Pathways",
        help="Spark application name (default: HTN_Treatment_Pathways)",
    )

    parser.add_argument(
        "--spark-master",
        default="local[*]",
        help="Spark master URL (default: local[*])",
    )

    return parser.parse_args()


def create_drug_concept_mapping(spark):
    """Create drug concept mapping for HTN medications - exact query from original."""
    drug_concept = spark.sql(
        """
    SELECT DISTINCT
        ancestor_concept_id,
        descendant_concept_id
    FROM
    (
        SELECT
            ancestor_concept_id,
            descendant_concept_id
        FROM concept_ancestor AS ca
        WHERE ca.ancestor_concept_id IN (21600381,21601461,21601560,21601664,21601744,21601782)
    ) a
    """
    )
    drug_concept.cache()
    drug_concept.createOrReplaceTempView("drug_concept")


def create_htn_index_cohort(spark):
    """Create HTN index cohort - exact query from original."""
    htn_index_cohort = spark.sql(
        """
    SELECT person_id, INDEX_DATE, COHORT_END_DATE, observation_period_start_date, observation_period_end_date
    FROM (
      SELECT ot.PERSON_ID, ot.INDEX_DATE, MIN(e.END_DATE) as COHORT_END_DATE, ot.OBSERVATION_PERIOD_START_DATE, ot.OBSERVATION_PERIOD_END_DATE,
             ROW_NUMBER() OVER (PARTITION BY ot.PERSON_ID ORDER BY ot.INDEX_DATE) as RowNumber
      FROM (
        SELECT dt.PERSON_ID, dt.DRUG_EXPOSURE_START_DATE as index_date, op.OBSERVATION_PERIOD_START_DATE, op.OBSERVATION_PERIOD_END_DATE
        FROM (
          SELECT de.PERSON_ID, de.DRUG_CONCEPT_ID, de.DRUG_EXPOSURE_START_DATE
          FROM (
            SELECT d.PERSON_ID, d.DRUG_CONCEPT_ID, d.DRUG_EXPOSURE_START_DATE,
                   COALESCE(d.DRUG_EXPOSURE_END_DATE, DATE_ADD(d.DRUG_EXPOSURE_START_DATE, d.DAYS_SUPPLY), DATE_ADD(d.DRUG_EXPOSURE_START_DATE, 1)) as DRUG_EXPOSURE_END_DATE,
                   ROW_NUMBER() OVER (PARTITION BY d.PERSON_ID ORDER BY d.DRUG_EXPOSURE_START_DATE) as RowNumber
            FROM (SELECT * FROM DRUG_EXPOSURE WHERE visit_occurrence_id IS NOT NULL) d
            JOIN drug_concept ca
              ON d.DRUG_CONCEPT_ID = ca.DESCENDANT_CONCEPT_ID AND ca.ANCESTOR_CONCEPT_ID IN (21600381, 21601461, 21601560, 21601664, 21601744, 21601782)
          ) de
          JOIN PERSON p ON p.PERSON_ID = de.PERSON_ID
          WHERE de.RowNumber = 1
        ) dt
        JOIN observation_period op
          ON op.PERSON_ID = dt.PERSON_ID AND (dt.DRUG_EXPOSURE_START_DATE BETWEEN op.OBSERVATION_PERIOD_START_DATE AND op.OBSERVATION_PERIOD_END_DATE)
        WHERE DATE_ADD(op.OBSERVATION_PERIOD_START_DATE, 365) <= dt.DRUG_EXPOSURE_START_DATE
          AND DATE_ADD(dt.DRUG_EXPOSURE_START_DATE, 1095) <= op.OBSERVATION_PERIOD_END_DATE
      ) ot
      JOIN (
        SELECT PERSON_ID, DATE_ADD(EVENT_DATE, -31) as END_DATE
        FROM (
          SELECT PERSON_ID, EVENT_DATE, EVENT_TYPE, START_ORDINAL,
                 ROW_NUMBER() OVER (PARTITION BY PERSON_ID ORDER BY EVENT_DATE, EVENT_TYPE) AS EVENT_ORDINAL,
                 MAX(START_ORDINAL) OVER (PARTITION BY PERSON_ID ORDER BY EVENT_DATE, EVENT_TYPE ROWS UNBOUNDED PRECEDING) as STARTS
          FROM (
            SELECT PERSON_ID, DRUG_EXPOSURE_START_DATE AS EVENT_DATE, 1 as EVENT_TYPE,
                   ROW_NUMBER() OVER (PARTITION BY PERSON_ID ORDER BY DRUG_EXPOSURE_START_DATE) as START_ORDINAL
            FROM (
              SELECT d.PERSON_ID, d.DRUG_CONCEPT_ID, d.DRUG_EXPOSURE_START_DATE,
                     COALESCE(d.DRUG_EXPOSURE_END_DATE, DATE_ADD(d.DRUG_EXPOSURE_START_DATE, d.DAYS_SUPPLY), DATE_ADD(d.DRUG_EXPOSURE_START_DATE, 1)) as DRUG_EXPOSURE_END_DATE,
                     ROW_NUMBER() OVER (PARTITION BY d.PERSON_ID ORDER BY DRUG_EXPOSURE_START_DATE) as RowNumber
              FROM (SELECT * FROM DRUG_EXPOSURE WHERE visit_occurrence_id IS NOT NULL) d
              JOIN drug_concept ca
                ON d.DRUG_CONCEPT_ID = ca.DESCENDANT_CONCEPT_ID AND ca.ANCESTOR_CONCEPT_ID IN (21600381, 21601461, 21601560, 21601664, 21601744, 21601782)
            ) cteExposureData
            UNION ALL
            SELECT PERSON_ID, DATE_ADD(DRUG_EXPOSURE_END_DATE, 31), 0 as EVENT_TYPE, NULL
            FROM (
              SELECT d.PERSON_ID, d.DRUG_CONCEPT_ID, d.DRUG_EXPOSURE_START_DATE,
                     COALESCE(d.DRUG_EXPOSURE_END_DATE, DATE_ADD(d.DRUG_EXPOSURE_START_DATE, d.DAYS_SUPPLY), DATE_ADD(d.DRUG_EXPOSURE_START_DATE, 1)) as DRUG_EXPOSURE_END_DATE,
                     ROW_NUMBER() OVER (PARTITION BY d.PERSON_ID ORDER BY DRUG_EXPOSURE_START_DATE) as RowNumber
              FROM (SELECT * FROM DRUG_EXPOSURE WHERE visit_occurrence_id IS NOT NULL) d
              JOIN drug_concept ca
                ON d.DRUG_CONCEPT_ID = ca.DESCENDANT_CONCEPT_ID AND ca.ANCESTOR_CONCEPT_ID IN (21600381, 21601461, 21601560, 21601664, 21601744, 21601782)
            ) cteExposureData
          ) RAWDATA
        ) E
        WHERE 2 * E.STARTS - E.EVENT_ORDINAL = 0
      ) e ON e.PERSON_ID = ot.PERSON_ID AND e.END_DATE >= ot.INDEX_DATE
      GROUP BY ot.PERSON_ID, ot.INDEX_DATE, ot.OBSERVATION_PERIOD_START_DATE, ot.OBSERVATION_PERIOD_END_DATE
    ) r
    WHERE r.RowNumber = 1
    """
    )

    htn_index_cohort.cache()
    htn_index_cohort.createOrReplaceTempView("htn_index_cohort")


def create_htn_e0(spark):
    """Create HTN_E0 - exact query from original."""
    HTN_E0 = spark.sql(
        """
    SELECT ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    FROM htn_index_cohort ip
    LEFT JOIN (
      SELECT co.PERSON_ID, co.CONDITION_CONCEPT_ID
      FROM condition_occurrence co
      JOIN htn_index_cohort ip ON co.PERSON_ID = ip.PERSON_ID
      JOIN drug_concept ca ON co.CONDITION_CONCEPT_ID = ca.DESCENDANT_CONCEPT_ID AND ca.ANCESTOR_CONCEPT_ID IN (444094)
      WHERE co.CONDITION_START_DATE BETWEEN ip.OBSERVATION_PERIOD_START_DATE AND ip.OBSERVATION_PERIOD_END_DATE
    ) dt ON dt.PERSON_ID = ip.PERSON_ID
    GROUP BY ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    HAVING COUNT(dt.CONDITION_CONCEPT_ID) <= 0
    """
    )

    HTN_E0.cache()
    HTN_E0.createOrReplaceTempView("HTN_E0")


def create_htn_t0(spark):
    """Create HTN_T0 - exact query from original."""
    HTN_T0 = spark.sql(
        """
    SELECT ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    FROM htn_index_cohort ip
    LEFT JOIN (
      SELECT de.PERSON_ID, de.DRUG_CONCEPT_ID
      FROM (SELECT * FROM DRUG_EXPOSURE WHERE visit_occurrence_id IS NOT NULL) de
      JOIN htn_index_cohort ip ON de.PERSON_ID = ip.PERSON_ID
      JOIN drug_concept ca ON de.DRUG_CONCEPT_ID = ca.DESCENDANT_CONCEPT_ID AND ca.ANCESTOR_CONCEPT_ID IN (21600381, 21601461, 21601560, 21601664, 21601744, 21601782)
      WHERE de.DRUG_EXPOSURE_START_DATE BETWEEN ip.OBSERVATION_PERIOD_START_DATE AND ip.OBSERVATION_PERIOD_END_DATE
        AND de.DRUG_EXPOSURE_START_DATE BETWEEN ip.OBSERVATION_PERIOD_START_DATE AND DATE_ADD(ip.INDEX_DATE, -1)
    ) dt ON dt.PERSON_ID = ip.PERSON_ID
    GROUP BY ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    HAVING COUNT(dt.DRUG_CONCEPT_ID) <= 0
    """
    )

    HTN_T0.createOrReplaceTempView("HTN_T0")


def create_htn_t1(spark):
    """Create HTN_T1 - exact query from original."""
    HTN_T1 = spark.sql(
        """
    SELECT ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    FROM htn_index_cohort ip
    LEFT JOIN (
      SELECT ce.PERSON_ID, ce.CONDITION_CONCEPT_ID
      FROM CONDITION_ERA ce
      JOIN htn_index_cohort ip ON ce.PERSON_ID = ip.PERSON_ID
      JOIN concept_ancestor ca ON ce.CONDITION_CONCEPT_ID = ca.DESCENDANT_CONCEPT_ID AND ca.ANCESTOR_CONCEPT_ID IN (316866)
      WHERE ce.CONDITION_ERA_START_DATE BETWEEN ip.OBSERVATION_PERIOD_START_DATE AND ip.OBSERVATION_PERIOD_END_DATE
    ) ct ON ct.PERSON_ID = ip.PERSON_ID
    GROUP BY ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    HAVING COUNT(ct.CONDITION_CONCEPT_ID) >= 1
    """
    )

    HTN_T1.createOrReplaceTempView("HTN_T1")


def create_htn_t2(spark):
    """Create HTN_T2 - exact query from original."""
    HTN_T2 = spark.sql(
        """
    SELECT ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    FROM htn_index_cohort ip
    LEFT JOIN (
      SELECT de.PERSON_ID, de.DRUG_CONCEPT_ID
      FROM (
        SELECT *
        FROM DRUG_EXPOSURE
        WHERE visit_occurrence_id IS NOT NULL
      ) de
      JOIN htn_index_cohort ip ON de.PERSON_ID = ip.PERSON_ID
      JOIN drug_concept ca ON de.DRUG_CONCEPT_ID = ca.DESCENDANT_CONCEPT_ID AND ca.ANCESTOR_CONCEPT_ID IN (21600381, 21601461, 21601560, 21601664, 21601744, 21601782)
      WHERE de.DRUG_EXPOSURE_START_DATE BETWEEN ip.OBSERVATION_PERIOD_START_DATE AND ip.OBSERVATION_PERIOD_END_DATE
        AND de.DRUG_EXPOSURE_START_DATE BETWEEN DATE_ADD(ip.INDEX_DATE, 121) AND DATE_ADD(ip.INDEX_DATE, 240)
    ) dt ON dt.PERSON_ID = ip.PERSON_ID
    GROUP BY ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    HAVING COUNT(dt.DRUG_CONCEPT_ID) >= 1
    """
    )

    HTN_T2.createOrReplaceTempView("HTN_T2")


def create_htn_t3(spark):
    """Create HTN_T3 - exact query from original."""
    HTN_T3 = spark.sql(
        """
    SELECT ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    FROM htn_index_cohort ip
    LEFT JOIN (
      SELECT de.PERSON_ID, de.DRUG_CONCEPT_ID
      FROM (
        SELECT *
        FROM DRUG_EXPOSURE
        WHERE visit_occurrence_id IS NOT NULL
      ) de
      JOIN htn_index_cohort ip ON de.PERSON_ID = ip.PERSON_ID
      JOIN drug_concept ca ON de.DRUG_CONCEPT_ID = ca.DESCENDANT_CONCEPT_ID AND ca.ANCESTOR_CONCEPT_ID IN (21600381, 21601461, 21601560, 21601664, 21601744, 21601782)
      WHERE de.DRUG_EXPOSURE_START_DATE BETWEEN ip.OBSERVATION_PERIOD_START_DATE AND ip.OBSERVATION_PERIOD_END_DATE
        AND de.DRUG_EXPOSURE_START_DATE BETWEEN DATE_ADD(ip.INDEX_DATE, 241) AND DATE_ADD(ip.INDEX_DATE, 360)
    ) dt ON dt.PERSON_ID = ip.PERSON_ID
    GROUP BY ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    HAVING COUNT(dt.DRUG_CONCEPT_ID) >= 1
    """
    )

    HTN_T3.createOrReplaceTempView("HTN_T3")


def create_htn_t4(spark):
    """Create HTN_T4 - exact query from original."""
    HTN_T4 = spark.sql(
        """
    SELECT ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    FROM htn_index_cohort ip
    LEFT JOIN (
      SELECT de.PERSON_ID, de.DRUG_CONCEPT_ID
      FROM (
        SELECT *
        FROM DRUG_EXPOSURE
        WHERE visit_occurrence_id IS NOT NULL
      ) de
      JOIN htn_index_cohort ip ON de.PERSON_ID = ip.PERSON_ID
      JOIN drug_concept ca ON de.DRUG_CONCEPT_ID = ca.DESCENDANT_CONCEPT_ID AND ca.ANCESTOR_CONCEPT_ID IN (21600381, 21601461, 21601560, 21601664, 21601744, 21601782)
      WHERE de.DRUG_EXPOSURE_START_DATE BETWEEN ip.OBSERVATION_PERIOD_START_DATE AND ip.OBSERVATION_PERIOD_END_DATE
        AND de.DRUG_EXPOSURE_START_DATE BETWEEN DATE_ADD(ip.INDEX_DATE, 361) AND DATE_ADD(ip.INDEX_DATE, 480)
    ) dt ON dt.PERSON_ID = ip.PERSON_ID
    GROUP BY ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    HAVING COUNT(dt.DRUG_CONCEPT_ID) >= 1
    """
    )

    HTN_T4.createOrReplaceTempView("HTN_T4")


def create_htn_t5(spark):
    """Create HTN_T5 - exact query from original."""
    HTN_T5 = spark.sql(
        """
    SELECT ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    FROM htn_index_cohort ip
    LEFT JOIN (
      SELECT de.PERSON_ID, de.DRUG_CONCEPT_ID
      FROM (
        SELECT *
        FROM DRUG_EXPOSURE
        WHERE visit_occurrence_id IS NOT NULL
      ) de
      JOIN htn_index_cohort ip ON de.PERSON_ID = ip.PERSON_ID
      JOIN drug_concept ca ON de.DRUG_CONCEPT_ID = ca.DESCENDANT_CONCEPT_ID AND ca.ANCESTOR_CONCEPT_ID IN (21600381, 21601461, 21601560, 21601664, 21601744, 21601782)
      WHERE de.DRUG_EXPOSURE_START_DATE BETWEEN ip.OBSERVATION_PERIOD_START_DATE AND ip.OBSERVATION_PERIOD_END_DATE
        AND de.DRUG_EXPOSURE_START_DATE BETWEEN DATE_ADD(ip.INDEX_DATE, 481) AND DATE_ADD(ip.INDEX_DATE, 600)
    ) dt ON dt.PERSON_ID = ip.PERSON_ID
    GROUP BY ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    HAVING COUNT(dt.DRUG_CONCEPT_ID) >= 1
    """
    )

    HTN_T5.createOrReplaceTempView("HTN_T5")


def create_htn_t6(spark):
    """Create HTN_T6 - exact query from original."""
    HTN_T6 = spark.sql(
        """
    SELECT ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    FROM htn_index_cohort ip
    LEFT JOIN (
      SELECT de.PERSON_ID, de.DRUG_CONCEPT_ID
      FROM (
        SELECT *
        FROM DRUG_EXPOSURE
        WHERE visit_occurrence_id IS NOT NULL
      ) de
      JOIN htn_index_cohort ip ON de.PERSON_ID = ip.PERSON_ID
      JOIN drug_concept ca ON de.DRUG_CONCEPT_ID = ca.DESCENDANT_CONCEPT_ID AND ca.ANCESTOR_CONCEPT_ID IN (21600381, 21601461, 21601560, 21601664, 21601744, 21601782)
      WHERE de.DRUG_EXPOSURE_START_DATE BETWEEN ip.OBSERVATION_PERIOD_START_DATE AND ip.OBSERVATION_PERIOD_END_DATE
        AND de.DRUG_EXPOSURE_START_DATE BETWEEN DATE_ADD(ip.INDEX_DATE, 601) AND DATE_ADD(ip.INDEX_DATE, 720)
    ) dt ON dt.PERSON_ID = ip.PERSON_ID
    GROUP BY ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    HAVING COUNT(dt.DRUG_CONCEPT_ID) >= 1
    """
    )

    HTN_T6.createOrReplaceTempView("HTN_T6")


def create_htn_t7(spark):
    """Create HTN_T7 - exact query from original."""
    HTN_T7 = spark.sql(
        """
    SELECT ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    FROM htn_index_cohort ip
    LEFT JOIN (
      SELECT de.PERSON_ID, de.DRUG_CONCEPT_ID
      FROM (
        SELECT *
        FROM DRUG_EXPOSURE
        WHERE visit_occurrence_id IS NOT NULL
      ) de
      JOIN htn_index_cohort ip ON de.PERSON_ID = ip.PERSON_ID
      JOIN drug_concept ca ON de.DRUG_CONCEPT_ID = ca.DESCENDANT_CONCEPT_ID AND ca.ANCESTOR_CONCEPT_ID IN (21600381, 21601461, 21601560, 21601664, 21601744, 21601782)
      WHERE de.DRUG_EXPOSURE_START_DATE BETWEEN ip.OBSERVATION_PERIOD_START_DATE AND ip.OBSERVATION_PERIOD_END_DATE
        AND de.DRUG_EXPOSURE_START_DATE BETWEEN DATE_ADD(ip.INDEX_DATE, 721) AND DATE_ADD(ip.INDEX_DATE, 840)
    ) dt ON dt.PERSON_ID = ip.PERSON_ID
    GROUP BY ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    HAVING COUNT(dt.DRUG_CONCEPT_ID) >= 1
    """
    )

    HTN_T7.createOrReplaceTempView("HTN_T7")


def create_htn_t8(spark):
    """Create HTN_T8 - exact query from original."""
    HTN_T8 = spark.sql(
        """
    SELECT ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    FROM htn_index_cohort ip
    LEFT JOIN (
      SELECT de.PERSON_ID, de.DRUG_CONCEPT_ID
      FROM (
        SELECT *
        FROM DRUG_EXPOSURE
        WHERE visit_occurrence_id IS NOT NULL
      ) de
      JOIN htn_index_cohort ip ON de.PERSON_ID = ip.PERSON_ID
      JOIN drug_concept ca ON de.DRUG_CONCEPT_ID = ca.DESCENDANT_CONCEPT_ID AND ca.ANCESTOR_CONCEPT_ID IN (21600381, 21601461, 21601560, 21601664, 21601744, 21601782)
      WHERE de.DRUG_EXPOSURE_START_DATE BETWEEN ip.OBSERVATION_PERIOD_START_DATE AND ip.OBSERVATION_PERIOD_END_DATE
        AND de.DRUG_EXPOSURE_START_DATE BETWEEN DATE_ADD(ip.INDEX_DATE, 841) AND DATE_ADD(ip.INDEX_DATE, 960)
    ) dt ON dt.PERSON_ID = ip.PERSON_ID
    GROUP BY ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    HAVING COUNT(dt.DRUG_CONCEPT_ID) >= 1
    """
    )

    HTN_T8.createOrReplaceTempView("HTN_T8")


def create_htn_t9(spark):
    """Create HTN_T9 - exact query from original."""
    HTN_T9 = spark.sql(
        """
    SELECT ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    FROM htn_index_cohort ip
    LEFT JOIN (
      SELECT de.PERSON_ID, de.DRUG_CONCEPT_ID
      FROM (
        SELECT *
        FROM DRUG_EXPOSURE
        WHERE visit_occurrence_id IS NOT NULL
      ) de
      JOIN htn_index_cohort ip ON de.PERSON_ID = ip.PERSON_ID
      JOIN drug_concept ca ON de.DRUG_CONCEPT_ID = ca.DESCENDANT_CONCEPT_ID AND ca.ANCESTOR_CONCEPT_ID IN (21600381, 21601461, 21601560, 21601664, 21601744, 21601782)
      WHERE de.DRUG_EXPOSURE_START_DATE BETWEEN ip.OBSERVATION_PERIOD_START_DATE AND ip.OBSERVATION_PERIOD_END_DATE
        AND de.DRUG_EXPOSURE_START_DATE BETWEEN DATE_ADD(ip.INDEX_DATE, 961) AND DATE_ADD(ip.INDEX_DATE, 1080)
    ) dt ON dt.PERSON_ID = ip.PERSON_ID
    GROUP BY ip.PERSON_ID, ip.INDEX_DATE, ip.COHORT_END_DATE
    HAVING COUNT(dt.DRUG_CONCEPT_ID) >= 1
    """
    )

    HTN_T9.createOrReplaceTempView("HTN_T9")


def create_htn_match_cohort(spark):
    """Create HTN_MatchCohort - exact query from original."""
    HTN_MatchCohort = spark.sql(
        """
    SELECT c.person_id, c.index_date, c.cohort_end_date, c.observation_period_start_date, c.observation_period_end_date
    FROM HTN_Index_Cohort C
    INNER JOIN (
      SELECT INDEX_DATE, COHORT_END_DATE, PERSON_ID
      FROM (
        SELECT INDEX_DATE, COHORT_END_DATE, PERSON_ID FROM HTN_E0
        INTERSECT
        SELECT INDEX_DATE, COHORT_END_DATE, PERSON_ID FROM HTN_T0
        INTERSECT
        SELECT INDEX_DATE, COHORT_END_DATE, PERSON_ID FROM HTN_T1
        INTERSECT
        SELECT INDEX_DATE, COHORT_END_DATE, PERSON_ID FROM HTN_T2
        INTERSECT
        SELECT INDEX_DATE, COHORT_END_DATE, PERSON_ID FROM HTN_T3
        INTERSECT
        SELECT INDEX_DATE, COHORT_END_DATE, PERSON_ID FROM HTN_T4
        INTERSECT
        SELECT INDEX_DATE, COHORT_END_DATE, PERSON_ID FROM HTN_T5
        INTERSECT
        SELECT INDEX_DATE, COHORT_END_DATE, PERSON_ID FROM HTN_T6
        INTERSECT
        SELECT INDEX_DATE, COHORT_END_DATE, PERSON_ID FROM HTN_T7
        INTERSECT
        SELECT INDEX_DATE, COHORT_END_DATE, PERSON_ID FROM HTN_T8
        INTERSECT
        SELECT INDEX_DATE, COHORT_END_DATE, PERSON_ID FROM HTN_T9
      ) TopGroup
    ) I
    ON C.PERSON_ID = I.PERSON_ID
    AND c.index_date = i.index_date
    """
    )

    return HTN_MatchCohort


def main():
    args = parse_arguments()

    # Initialize Spark
    spark = SparkSession.builder.appName(f"HTN treatment pathway").getOrCreate()

    try:
        # Load the source OMOP tables
        person = spark.read.parquet(os.path.join(args.omop_folder, "person"))
        visit_occurrence = spark.read.parquet(
            os.path.join(args.omop_folder, "visit_occurrence")
        )
        condition_occurrence = spark.read.parquet(
            os.path.join(args.omop_folder, "condition_occurrence")
        )
        procedure_occurrence = spark.read.parquet(
            os.path.join(args.omop_folder, "procedure_occurrence")
        )
        drug_exposure = spark.read.parquet(
            os.path.join(args.omop_folder, "drug_exposure")
        )
        observation_period = spark.read.parquet(
            os.path.join(args.omop_folder, "observation_period")
        )
        condition_era = spark.read.parquet(
            os.path.join(args.omop_folder, "condition_era")
        )

        print(f"person: {visit_occurrence.select('person_id').distinct().count()}")
        print(f"visit_occurrence: {visit_occurrence.count()}")
        print(f"condition_occurrence: {condition_occurrence.count()}")
        print(f"procedure_occurrence: {procedure_occurrence.count()}")
        print(f"drug_exposure: {drug_exposure.count()}")
        print(f"observation_period: {observation_period.count()}")

        concept = spark.read.parquet(os.path.join(args.omop_folder, "concept"))
        concept_ancestor = spark.read.parquet(
            os.path.join(args.omop_folder, "concept_ancestor")
        )

        # Create temporary views
        person.createOrReplaceTempView("person")
        visit_occurrence.createOrReplaceTempView("visit_occurrence")
        condition_occurrence.createOrReplaceTempView("condition_occurrence")
        procedure_occurrence.createOrReplaceTempView("procedure_occurrence")
        drug_exposure.createOrReplaceTempView("drug_exposure")
        observation_period.createOrReplaceTempView("observation_period")
        condition_era.createOrReplaceTempView("condition_era")

        concept_ancestor.createOrReplaceTempView("concept_ancestor")
        concept.createOrReplaceTempView("concept")

        # Create drug concept mapping
        create_drug_concept_mapping(spark)

        # Create HTN index cohort
        create_htn_index_cohort(spark)

        # Create all cohorts in sequence
        create_htn_e0(spark)
        create_htn_t0(spark)
        create_htn_t1(spark)
        create_htn_t2(spark)
        create_htn_t3(spark)
        create_htn_t4(spark)
        create_htn_t5(spark)
        create_htn_t6(spark)
        create_htn_t7(spark)
        create_htn_t8(spark)
        create_htn_t9(spark)

        # Create final cohort
        htn_match_cohort = create_htn_match_cohort(spark)

        # # Save results
        # if not os.path.exists(args.output_folder):
        #     os.makedirs(args.output_folder)
        #
        # output_path = os.path.join(args.output_folder, "htn_match_cohort")
        # htn_match_cohort.write.mode("overwrite").parquet(output_path)
        #
        # # Read back and count
        # saved_cohort = spark.read.parquet(output_path)
        final_count = htn_match_cohort.count()
        print(f"Final cohort count: {final_count}")

        print("Analysis completed successfully!")

    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)
    finally:
        spark.stop()


if __name__ == "__main__":
    main()



import os
import shutil
import sys
import tempfile
import unittest
from pathlib import Path

import pandas as pd
from datasets import disable_caching

from cehrgpt.generation.generate_batch_hf_gpt_sequence import create_arg_parser
from cehrgpt.generation.generate_batch_hf_gpt_sequence import main as generate_main
from cehrgpt.models.pretrained_embeddings import (
    PRETRAINED_EMBEDDING_CONCEPT_FILE_NAME,
    PRETRAINED_EMBEDDING_VECTOR_FILE_NAME,
)
from cehrgpt.runners.hf_cehrgpt_pretrain_runner import main as train_main

disable_caching()
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
os.environ["WANDB_MODE"] = "disabled"
os.environ["TRANSFORMERS_VERBOSITY"] = "info"


class HfCehrGptRunnerIntegrationTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Get the root folder of the project
        root_folder = Path(os.path.abspath(__file__)).parent.parent.parent.parent
        cls.data_folder = os.path.join(root_folder, "sample_data", "pretrain")
        cls.pretrained_embedding_folder = os.path.join(
            root_folder, "sample_data", "pretrained_embeddings"
        )
        cls.motor_vocab_dir = os.path.join(root_folder, "sample_data", "omop_vocab")
        # Create a temporary directory to store model and tokenizer
        cls.temp_dir = tempfile.mkdtemp()
        cls.model_folder_path = os.path.join(cls.temp_dir, "model")
        Path(cls.model_folder_path).mkdir(parents=True, exist_ok=True)
        cls.dataset_prepared_path = os.path.join(cls.temp_dir, "dataset_prepared_path")
        Path(cls.dataset_prepared_path).mkdir(parents=True, exist_ok=True)
        cls.generation_folder_path = os.path.join(cls.temp_dir, "generation")
        Path(cls.generation_folder_path).mkdir(parents=True, exist_ok=True)
        for file_name in [
            PRETRAINED_EMBEDDING_CONCEPT_FILE_NAME,
            PRETRAINED_EMBEDDING_VECTOR_FILE_NAME,
        ]:
            shutil.copy(
                os.path.join(cls.pretrained_embedding_folder, file_name),
                os.path.join(cls.model_folder_path, file_name),
            )

    @classmethod
    def tearDownClass(cls):
        # Remove the temporary directory
        shutil.rmtree(cls.temp_dir)

    def test_1_train_model(self):
        sys.argv = [
            "hf_cehrgpt_pretraining_runner.py",
            "--model_name_or_path",
            self.model_folder_path,
            "--tokenizer_name_or_path",
            self.model_folder_path,
            "--output_dir",
            self.model_folder_path,
            "--vocab_dir",
            self.motor_vocab_dir,
            "--data_folder",
            self.data_folder,
            "--dataset_prepared_path",
            self.dataset_prepared_path,
            "--pretrained_embedding_path",
            self.model_folder_path,
            "--num_train_epochs",
            "1",
            "--save_steps",
            "1",
            "--save_strategy",
            "steps",
            "--eval_steps",
            "1",
            "--evaluation_strategy",
            "steps",
            "--hidden_size",
            "32",
            "--max_position_embeddings",
            "32",
            "--use_sub_time_tokenization",
            "false",
            "--include_ttv_prediction",
            "false",
            "--include_values",
            "true",
            "--lab_token_penalty",
            "true",
            "--entropy_penalty",
            "true",
            "--validation_split_num",
            "10",
            "--streaming",
            "false",
            "--use_early_stopping",
            "false",
            "--sample_packing",
            "true",
            "--sample_packing",
            "true",
            "--max_tokens_per_batch",
            "512",
            "--report_to",
            "none",
            "--include_motor_time_to_event",
            "true",
            "--exclude_position_ids",
            "true",
            "--apply_rotary",
            "true",
        ]
        train_main()
        # Teacher force the prompt to consist of [year][age][gender][race][VS] then inject the random vector before [VS]


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from scipy.interpolate import UnivariateSpline

from cehrgpt.models.tokenization_hf_cehrgpt import create_sample_from_bins


class TestCreateSampleFromBins(unittest.TestCase):
    def setUp(self):
        # Create a set of bins with splines
        x = np.linspace(0, 10, 100)
        y = np.sin(x)
        spline = UnivariateSpline(x, y, s=1)
        self.bins = [{"spline": spline} for _ in range(5)]

    def test_sample_size(self):
        sample_size = 10_000
        samples = create_sample_from_bins(self.bins, sample_size)
        self.assertEqual(len(samples), sample_size)

    def test_empty_bins(self):
        samples = create_sample_from_bins([], 10_000)
        self.assertEqual(len(samples), 0)

    def test_zero_sample_size(self):
        samples = create_sample_from_bins(self.bins, 0)
        self.assertEqual(len(samples), 0)

    def test_sample_distribution(self):
        """Check if samples respect the distribution from the spline by checking mean close to expected."""
        x = np.linspace(0, 10, 10)
        y = np.sin(x)
        spline = UnivariateSpline(x, y, s=1)
        bins = [{"spline": spline}]
        samples = create_sample_from_bins(bins, 1000)
        expected_mean = np.mean(spline(x))
        samples_mean = np.mean(samples)
        self.assertAlmostEqual(samples_mean, expected_mean, delta=0.1)


if __name__ == "__main__":
    unittest.main()



from typing import Optional, Union

import torch
from datasets import Dataset
from torch.utils.data import DataLoader
from transformers import Trainer
from transformers.trainer_utils import has_length, seed_worker
from transformers.utils import import_utils, logging

from cehrgpt.data.sample_packing_sampler import SamplePackingBatchSampler

DEFAULT_MAX_TOKENS_PER_BATCH = 16384

LOG = logging.get_logger("transformers")


class SamplePackingTrainer(Trainer):
    def __init__(self, *args, **kwargs):
        if "max_tokens_per_batch" in kwargs:
            self.max_tokens_per_batch = kwargs.pop("max_tokens_per_batch")
            LOG.info("max_tokens_per_batch: %s", self.max_tokens_per_batch)
        else:
            self.max_tokens_per_batch = DEFAULT_MAX_TOKENS_PER_BATCH
            LOG.info(
                "max_tokens_per_batch is not provided to SamplePackingTrainer and will default to %s",
                DEFAULT_MAX_TOKENS_PER_BATCH,
            )

        if "max_position_embeddings" in kwargs:
            self.max_position_embeddings = kwargs.pop("max_position_embeddings")
            LOG.info("max_position_embeddings: %s", self.max_position_embeddings)
        else:
            self.max_position_embeddings = self.max_tokens_per_batch
            LOG.info(
                "max_position_embeddings is not provided to SamplePackingTrainer and will default to %s",
                self.max_tokens_per_batch,
            )

        self.negative_sampling_probability = kwargs.pop(
            "negative_sampling_probability", None
        )
        if self.negative_sampling_probability:
            LOG.info(
                "negative_sampling_probability: %s", self.negative_sampling_probability
            )
        self.train_lengths = kwargs.pop("train_lengths", None)
        self.validation_lengths = kwargs.pop("validation_lengths", None)
        super().__init__(*args, **kwargs)
        self.accelerator.even_batches = False

    def num_examples(self, dataloader: DataLoader) -> int:
        if has_length(dataloader):
            return len(dataloader)
        raise RuntimeError("DataLoader in SamplePackingTrainer must have length")

    def get_train_dataloader(self) -> DataLoader:
        """Returns the training dataloader with our custom batch sampler."""
        train_dataset = self.train_dataset

        if self.train_lengths is None:
            LOG.info("Started computing lengths for the train dataset")
            # Calculate lengths of all sequences in dataset
            if "num_of_concepts" in train_dataset.column_names:
                lengths = train_dataset["num_of_concepts"]
            else:
                lengths = [
                    len(sample["input_ids"])
                    for sample in train_dataset.select_columns("input_ids")
                ]

            LOG.info("Finished computing lengths for the train dataset")
        else:
            lengths = self.train_lengths

        data_collator = self.data_collator
        if import_utils.is_datasets_available() and isinstance(train_dataset, Dataset):
            train_dataset = self._remove_unused_columns(
                train_dataset, description="training"
            )
        else:
            data_collator = self._get_collator_with_removed_columns(
                data_collator, description="training"
            )

        labels = None
        if (
            self.negative_sampling_probability is not None
            and "classifier_label" in train_dataset.column_names
        ):
            labels = train_dataset["classifier_label"]

        # Create our custom batch sampler
        batch_sampler = SamplePackingBatchSampler(
            lengths=lengths,
            max_tokens_per_batch=self.max_tokens_per_batch,
            max_position_embeddings=self.max_position_embeddings,
            drop_last=self.args.dataloader_drop_last,
            seed=self.args.seed,
            negative_sampling_probability=self.negative_sampling_probability,
            labels=labels,
        )
        dataloader_params = {
            "collate_fn": data_collator,
            "num_workers": self.args.dataloader_num_workers,
            "pin_memory": self.args.dataloader_pin_memory,
            "persistent_workers": self.args.dataloader_persistent_workers,
            "batch_sampler": batch_sampler,
        }
        if not isinstance(train_dataset, torch.utils.data.IterableDataset):
            dataloader_params["drop_last"] = self.args.dataloader_drop_last
            dataloader_params["worker_init_fn"] = seed_worker
            dataloader_params["prefetch_factor"] = self.args.dataloader_prefetch_factor

        return self.accelerator.prepare(DataLoader(train_dataset, **dataloader_params))

    def get_eval_dataloader(
        self, eval_dataset: Optional[Union[str, Dataset]] = None
    ) -> DataLoader:
        """
        Returns the evaluation [`~torch.utils.data.DataLoader`].

        Subclass and override this method if you want to inject some custom behavior.

        Args:
            eval_dataset (`str` or `torch.utils.data.Dataset`, *optional*):
                If a `str`, will use `self.eval_dataset[eval_dataset]` as the evaluation dataset. If a `Dataset`, will override `self.eval_dataset` and must implement `__len__`. If it is a [`~datasets.Dataset`], columns not accepted by the `model.forward()` method are automatically removed.
        """
        if eval_dataset is None and self.eval_dataset is None:
            raise ValueError("Trainer: evaluation requires an eval_dataset.")

        # If we have persistent workers, don't do a fork bomb especially as eval datasets
        # don't change during training
        dataloader_key = eval_dataset if isinstance(eval_dataset, str) else "eval"
        if (
            hasattr(self, "_eval_dataloaders")
            and dataloader_key in self._eval_dataloaders
            and self.args.dataloader_persistent_workers
        ):
            return self.accelerator.prepare(self._eval_dataloaders[dataloader_key])

        eval_dataset = (
            self.eval_dataset[eval_dataset]
            if isinstance(eval_dataset, str)
            else eval_dataset if eval_dataset is not None else self.eval_dataset
        )

        if self.validation_lengths is None:
            LOG.info("Started computing lengths for the train dataset")
            # Calculate lengths of all sequences in dataset
            if "num_of_concepts" in eval_dataset.column_names:
                lengths = eval_dataset["num_of_concepts"]
            else:
                lengths = [len(sample["input_ids"]) for sample in eval_dataset]

            LOG.info("Finished computing lengths for the train dataset")
        else:
            lengths = self.validation_lengths

        data_collator = self.data_collator

        if import_utils.is_datasets_available() and isinstance(eval_dataset, Dataset):
            eval_dataset = self._remove_unused_columns(
                eval_dataset, description="evaluation"
            )
        else:
            data_collator = self._get_collator_with_removed_columns(
                data_collator, description="evaluation"
            )

        # Create our custom batch sampler
        batch_sampler = SamplePackingBatchSampler(
            lengths=lengths,
            max_tokens_per_batch=self.max_tokens_per_batch,
            max_position_embeddings=self.max_position_embeddings,
            drop_last=self.args.dataloader_drop_last,
            seed=self.args.seed,
        )
        dataloader_params = {
            "collate_fn": data_collator,
            "num_workers": self.args.dataloader_num_workers,
            "pin_memory": self.args.dataloader_pin_memory,
            "persistent_workers": self.args.dataloader_persistent_workers,
            "batch_sampler": batch_sampler,
        }
        # accelerator.free_memory() will destroy the references, so
        # we need to store the non-prepared version
        eval_dataloader = DataLoader(eval_dataset, **dataloader_params)
        if self.args.dataloader_persistent_workers:
            if hasattr(self, "_eval_dataloaders"):
                self._eval_dataloaders[dataloader_key] = eval_dataloader
            else:
                self._eval_dataloaders = {dataloader_key: eval_dataloader}

        return self.accelerator.prepare(eval_dataloader)



from datetime import date
from typing import Any, Dict, List, Optional

import cehrbert_data.cohorts.spark_app_base
from cehrbert_data.cohorts.query_builder import QueryBuilder, QuerySpec
from cehrbert_data.cohorts.spark_app_base import BaseCohortBuilder
from cehrbert_data.utils.spark_utils import preprocess_domain_table


def custom_instantiate_dependencies(spark, input_folder, dependency_list):
    dependency_dict = dict()
    for domain_table_name in dependency_list:
        table = preprocess_domain_table(spark, input_folder, domain_table_name)
        table.createOrReplaceTempView(domain_table_name)
        dependency_dict[domain_table_name] = table
    return dependency_dict


# Monkeypatch the function
cehrbert_data.cohorts.spark_app_base.instantiate_dependencies = (
    custom_instantiate_dependencies
)


class OmopTableBuilder(BaseCohortBuilder):
    cohort_required_columns = []

    def __init__(
        self,
        query_builder: QueryBuilder,
        input_folder: str,
        output_folder: str,
        continue_job: bool = False,
    ):
        super().__init__(
            query_builder=query_builder,
            input_folder=input_folder,
            output_folder=output_folder,
            date_lower_bound="1900-01-01",
            date_upper_bound=date.today().strftime("%Y-%m-%d"),
            age_lower_bound=0,
            age_upper_bound=200,
            prior_observation_period=0,
            post_observation_period=0,
            continue_job=continue_job,
        )

    def build(self):
        # Check whether the cohort has been generated
        if self._continue_job and self.cohort_exists():
            return self
        cohort = self.create_cohort()
        cohort.write.mode("overwrite").parquet(self._output_data_folder)
        return self

    @staticmethod
    def create_omop_query_builder(
        input_folder: str,
        output_folder: str,
        table_name: str,
        query_template: str,
        dependency_list: List[str],
        query_parameters: Optional[Dict[str, Any]] = None,
        continue_job: bool = False,
    ):

        if query_parameters is None:
            query_parameters = dict()
        query = QuerySpec(
            table_name=table_name,
            query_template=query_template,
            parameters=query_parameters,
        )
        table_query_builder = QueryBuilder(
            cohort_name=table_name,
            dependency_list=dependency_list,
            query=query,
            ancestor_table_specs=[],
        )

        return OmopTableBuilder(
            query_builder=table_query_builder,
            input_folder=input_folder,
            output_folder=output_folder,
            continue_job=continue_job,
        )



import argparse

from cehrbert_data.utils.spark_utils import validate_table_names


def create_omop_argparse():
    parser = argparse.ArgumentParser(
        description="Spark application for creating OMOP table"
    )
    parser.add_argument(
        "--input_folder",
        dest="input_folder",
        action="store",
        help="The path for your input_folder where the raw data is",
        required=True,
    )
    parser.add_argument(
        "--output_folder",
        dest="output_folder",
        action="store",
        help="The path for your output_folder",
        required=True,
    )
    parser.add_argument(
        "--continue_job",
        dest="continue_job",
        action="store_true",
    )
    parser.add_argument(
        "--domain_table_list",
        dest="domain_table_list",
        nargs="+",
        action="store",
        help="The list of domain tables you want to download",
        type=validate_table_names,
        required=True,
    )
    return parser



import unittest

import numpy as np
from scipy.interpolate import UnivariateSpline

from cehrgpt.models.tokenization_hf_cehrgpt import (
    NA,
    UNKNOWN_BIN,
    NumericEventStatistics,
    create_numeric_concept_unit_mapping,
)


class TestNumericEventStatistics(unittest.TestCase):
    def setUp(self):
        self.spline = UnivariateSpline(list(range(10)), list(range(10)))
        self.spline_2 = UnivariateSpline(list(range(10)), list(range(100, 110)))
        # Mock lab stats data
        self.lab_stats = [
            {
                "concept_id": "concept_1",
                "unit": "unit_1",
                "mean": 10.0,
                "std": 2.0,
                "value_outlier_std": 3.0,
                "bins": [
                    {
                        "bin_index": 0,
                        "spline": self.spline,
                        "start_val": 0,
                        "end_val": 10,
                    }
                ],
                "count": 100,
            },
            {
                "concept_id": "concept_2",
                "unit": "unit_1",
                "mean": 20.0,
                "std": 5.0,
                "value_outlier_std": 2.5,
                "bins": [
                    {
                        "bin_index": 0,
                        "spline": self.spline_2,
                        "start_val": 100,
                        "end_val": 110,
                    }
                ],
                "count": 100,
            },
            {
                "concept_id": "concept_2",
                "unit": "unit_2",
                "mean": 15.0,
                "std": 3.0,
                "value_outlier_std": 2.0,
                "bins": [
                    {
                        "bin_index": 0,
                        "spline": self.spline,
                        "start_val": 0,
                        "end_val": 10,
                    }
                ],
                "count": 200,
            },
        ]
        # Create an instance of NumericEventStatistics
        self.numeric_event_statistics = NumericEventStatistics(self.lab_stats)

    def test_create_numeric_concept_unit_mapping(self):
        # Call the function
        concept_prob_mapping, concept_unit_mapping = (
            create_numeric_concept_unit_mapping(self.lab_stats)
        )

        # Check the concept_prob_mapping
        # For concept_1: Only one unit, so probability = 1.0
        self.assertEqual(concept_prob_mapping["concept_1"], [1.0])

        # For concept_2: Two units, unit_1 with count 100, unit_2 with count 200
        total_count_concept_2 = 100 + 200
        expected_probs_concept_2 = [
            100 / total_count_concept_2,
            200 / total_count_concept_2,
        ]
        self.assertEqual(concept_prob_mapping["concept_2"], expected_probs_concept_2)

        # Check the concept_unit_mapping
        self.assertEqual(concept_unit_mapping["concept_1"], ("unit_1",))
        self.assertEqual(concept_unit_mapping["concept_2"], ("unit_1", "unit_2"))

    def test_get_numeric_concept_ids(self):
        # Test for correct concept IDs
        expected_concept_ids = ["concept_1", "concept_2", "concept_2"]
        result = self.numeric_event_statistics.get_numeric_concept_ids()
        self.assertEqual(result, expected_concept_ids)

    def test_get_random_unit(self):
        # Test get_random_unit method for concept_1 (single unit)
        unit = self.numeric_event_statistics.get_random_unit("concept_1")
        self.assertEqual(unit, "unit_1")

        # Test get_random_unit method for concept_2 (multiple units)
        unit = self.numeric_event_statistics.get_random_unit("concept_2")
        self.assertIn(unit, ["unit_1", "unit_2"])

        # Test get_random_unit method for non-existent concept_3
        unit = self.numeric_event_statistics.get_random_unit("concept_3")
        self.assertEqual(unit, NA)

    def test_normalize(self):
        # Test normalization for concept_1, unit_1
        normalized_value = self.numeric_event_statistics.normalize(
            "concept_1", "unit_1", 9.0
        )
        # (12 - 10) / 2 = 1.0
        self.assertEqual(normalized_value, "BIN:0")

        # Test normalization with value beyond outlier bound
        normalized_value = self.numeric_event_statistics.normalize(
            "concept_1", "unit_1", 3.0
        )
        # Since (20 - 10) / 2 = 5.0, but it's clipped at value_outlier_std = 3.0
        self.assertEqual(normalized_value, "BIN:0")

        # Test normalization with a concept_id/unit pair not found
        normalized_value = self.numeric_event_statistics.normalize(
            "concept_2", "unit_1", 105.0
        )
        # Since the concept_id/unit doesn't exist, it should return the original value
        self.assertEqual(normalized_value, "BIN:0")

        # Test normalization with a concept_id/unit pair not found
        normalized_value = self.numeric_event_statistics.normalize(
            "concept_2", "unit_1", 9.9
        )
        # Since the concept_id/unit doesn't exist, it should return the original value
        self.assertEqual(normalized_value, UNKNOWN_BIN)

    def test_denormalize(self):
        # Mock np.random.choice to return a predictable unit
        np.random.choice = lambda *args, **kwargs: "unit_1"
        np.random.seed(42)
        expected_value = self.spline(
            np.random.uniform(self.spline.get_knots()[0], self.spline.get_knots()[-1])
        )
        np.random.seed(42)
        # Test denormalization for concept_1, unit_1
        denormalized_value, unit = self.numeric_event_statistics.denormalize(
            "concept_1", "BIN:0"
        )
        # value = 1.0 * 2.0 + 10.0 = 12.0
        self.assertEqual(denormalized_value, expected_value)
        self.assertEqual(unit, "unit_1")

        np.random.seed(42)
        expected_value = self.spline_2(
            np.random.uniform(
                self.spline_2.get_knots()[0], self.spline_2.get_knots()[-1]
            )
        )
        np.random.seed(42)
        # Test denormalization for concept_2, unit_1
        denormalized_value, unit = self.numeric_event_statistics.denormalize(
            "concept_2", "BIN:0"
        )
        self.assertEqual(denormalized_value, expected_value)
        self.assertEqual(unit, "unit_1")


# Entry point for running the tests
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
Depression Treatment Pathways Study Script.

Specialized script for analyzing depression treatment pathways using OMOP CDM data.
Uses the generalized treatment_pathway module with depression-specific parameters.
"""

import argparse
import sys

from cehrgpt.analysis.treatment_pathway.treatment_pathway import main as treatment_main

# Depression-specific configuration
DEPRESSION_CONFIG = {
    "study_name": "DEPRESSION",
    "drug_concepts": [21604686, 21500526],
    "target_conditions": [440383],
    "exclusion_conditions": [444094, 432876, 435783],
}


def parse_depression_arguments():
    """Parse command line arguments specific to depression study."""
    parser = argparse.ArgumentParser(
        description="Analyze depression treatment pathways using OMOP CDM data"
    )

    parser.add_argument(
        "--omop-folder",
        required=True,
        help="Path to OMOP CDM data folder containing parquet files",
    )

    parser.add_argument(
        "--output-folder", required=True, help="Output folder for results"
    )

    parser.add_argument(
        "--app-name",
        default="Depression_Treatment_Pathways",
        help="Spark application name (default: Depression_Treatment_Pathways)",
    )

    parser.add_argument(
        "--spark-master",
        default="local[*]",
        help="Spark master URL (default: local[*])",
    )

    return parser.parse_args()


def main():
    """Main function that overwrites sys.argv and calls the generalized treatment pathway main."""
    depression_args = parse_depression_arguments()

    print("Depression Treatment Pathways Analysis")
    print("=" * 50)
    print(f"Study Configuration:")
    print(f"  - Study Name: {DEPRESSION_CONFIG['study_name']}")
    print(f"  - Drug Concepts: {DEPRESSION_CONFIG['drug_concepts']}")
    print(f"  - Target Conditions: {DEPRESSION_CONFIG['target_conditions']}")
    print(f"  - Exclusion Conditions: {DEPRESSION_CONFIG['exclusion_conditions']}")
    print(f"  - OMOP Folder: {depression_args.omop_folder}")
    print(f"  - Output Folder: {depression_args.output_folder}")
    print("=" * 50)

    # Overwrite sys.argv with the generalized treatment pathway arguments
    sys.argv = [
        "treatment_pathway.py",
        "--omop-folder",
        depression_args.omop_folder,
        "--output-folder",
        depression_args.output_folder,
        "--drug-concepts",
        ",".join(map(str, DEPRESSION_CONFIG["drug_concepts"])),
        "--target-conditions",
        ",".join(map(str, DEPRESSION_CONFIG["target_conditions"])),
        "--exclusion-conditions",
        ",".join(map(str, DEPRESSION_CONFIG["exclusion_conditions"])),
        "--app-name",
        depression_args.app_name,
        "--spark-master",
        depression_args.spark_master,
        "--study-name",
        DEPRESSION_CONFIG["study_name"],
    ]
    # Parse arguments using the generalized script's parser and call its main function
    treatment_main()


if __name__ == "__main__":
    main()



import logging
import os
import random
from datetime import datetime

import numpy as np
import pandas as pd

from cehrgpt.models.tokenization_hf_cehrgpt import CehrGptTokenizer

from .utils import (
    RANDOM_SEE,
    create_demographics,
    create_gender_encoder,
    create_race_encoder,
    create_vector_representations,
    find_match,
    find_match_self,
    scale_age,
)

logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger("NNAA")
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")


def main(args):
    LOG.info(f"Started loading tokenizer at {args.concept_tokenizer_path}")
    concept_tokenizer = CehrGptTokenizer.from_pretrained(args.concept_tokenizer_path)

    LOG.info(f"Started loading training data at {args.training_data_folder}")
    train_data = pd.read_parquet(args.training_data_folder)

    LOG.info(f"Started loading evaluation data at {args.evaluation_data_folder}")
    evaluation_data = pd.read_parquet(args.evaluation_data_folder)

    LOG.info(f"Started loading synthetic_data at {args.synthetic_data_folder}")
    synthetic_data = pd.read_parquet(args.synthetic_data_folder)

    LOG.info(
        "Started extracting the demographic information from the patient sequences"
    )
    train_data = create_demographics(train_data)
    evaluation_data = create_demographics(evaluation_data)
    synthetic_data = create_demographics(synthetic_data)

    LOG.info("Started rescaling age columns")
    train_data = scale_age(train_data)
    evaluation_data = scale_age(evaluation_data)
    synthetic_data = scale_age(synthetic_data)

    LOG.info("Started encoding gender")
    gender_encoder = create_gender_encoder(train_data, evaluation_data, synthetic_data)
    LOG.info("Completed encoding gender")

    LOG.info("Started encoding race")
    race_encoder = create_race_encoder(train_data, evaluation_data, synthetic_data)
    LOG.info("Completed encoding race")

    random.seed(RANDOM_SEE)
    metrics = []

    for i in range(args.n_iterations):
        LOG.info(f"Iteration {i} Started")
        train_data_sample = train_data.sample(args.num_of_samples)
        evaluation_data_sample = evaluation_data.sample(args.num_of_samples)
        synthetic_data_sample = synthetic_data.sample(args.num_of_samples)
        LOG.info(f"Iteration {i}: Started creating train vectors")
        train_vectors = create_vector_representations(
            train_data_sample, concept_tokenizer, gender_encoder, race_encoder
        )
        LOG.info(f"Iteration {i}: Started creating evaluation vectors")
        evaluation_vectors = create_vector_representations(
            evaluation_data_sample, concept_tokenizer, gender_encoder, race_encoder
        )
        LOG.info(f"Iteration {i}: Started creating synthetic vectors")
        synthetic_vectors = create_vector_representations(
            synthetic_data_sample, concept_tokenizer, gender_encoder, race_encoder
        )
        LOG.info(
            f"Iteration {i}: Started calculating the distances between synthetic and training vectors"
        )
        distance_train_TS = find_match(train_vectors, synthetic_vectors)
        distance_train_ST = find_match(synthetic_vectors, train_vectors)
        distance_train_TT = find_match_self(train_vectors, train_vectors)
        distance_train_SS = find_match_self(synthetic_vectors, synthetic_vectors)

        aa_train = (
            (
                np.sum(distance_train_TS > distance_train_TT)
                + np.sum(distance_train_ST > distance_train_SS)
            )
            / args.num_of_samples
            / 2
        )

        LOG.info(
            f"Iteration {i}: Started calculating the distances between synthetic and evaluation vectors"
        )
        distance_test_TS = find_match(evaluation_vectors, synthetic_vectors)
        distance_test_ST = find_match(synthetic_vectors, evaluation_vectors)
        distance_test_TT = find_match_self(evaluation_vectors, evaluation_vectors)
        distance_test_SS = find_match_self(synthetic_vectors, synthetic_vectors)

        aa_test = (
            (
                np.sum(distance_test_TS > distance_test_TT)
                + np.sum(distance_test_ST > distance_test_SS)
            )
            / args.num_of_samples
            / 2
        )

        privacy_loss = aa_test - aa_train
        metrics.append(privacy_loss)
        LOG.info(f"Iteration {i}: Privacy loss {privacy_loss}")

    results = {"NNAAE": metrics}

    current_time = datetime.now().strftime("%m-%d-%Y-%H-%M-%S")
    pd.DataFrame([results], columns=["NNAAE"]).to_parquet(
        os.path.join(
            args.metrics_folder, f"nearest_neighbor_inference_{current_time}.parquet"
        )
    )


def create_argparser():
    import argparse

    parser = argparse.ArgumentParser(
        description="Nearest Neighbor Inference Analysis Arguments using the GPT model"
    )
    parser.add_argument(
        "--training_data_folder",
        dest="training_data_folder",
        action="store",
        help="The path for where the training data folder",
        required=True,
    )
    parser.add_argument(
        "--evaluation_data_folder",
        dest="evaluation_data_folder",
        action="store",
        help="The path for where the evaluation data folder",
        required=True,
    )
    parser.add_argument(
        "--synthetic_data_folder",
        dest="synthetic_data_folder",
        action="store",
        help="The path for where the synthetic data folder",
        required=True,
    )
    parser.add_argument(
        "--concept_tokenizer_path",
        dest="concept_tokenizer_path",
        action="store",
        help="The path for where the concept tokenizer is located",
        required=True,
    )
    parser.add_argument(
        "--num_of_samples",
        dest="num_of_samples",
        action="store",
        type=int,
        required=False,
        default=5000,
    )
    parser.add_argument(
        "--n_iterations",
        dest="n_iterations",
        action="store",
        type=int,
        required=False,
        default=1,
    )
    parser.add_argument(
        "--metrics_folder",
        dest="metrics_folder",
        action="store",
        help="The folder that stores the metrics",
        required=False,
    )
    return parser


if __name__ == "__main__":
    main(create_argparser().parse_args())



import pickle
from pathlib import Path
from typing import Optional

import numpy as np

PRETRAINED_EMBEDDING_VECTOR_FILE_NAME = "pretrained_embedding_vectors.npy"
PRETRAINED_EMBEDDING_CONCEPT_FILE_NAME = "pretrained_embedding_concepts.pkl"


class PretrainedEmbeddings:
    """A class to handle pretrained embedding vectors and their associated concepts."""

    def __init__(self, model_folder: Optional[str]):
        if model_folder:
            model_path = Path(model_folder)
            self.vector_file = model_path / PRETRAINED_EMBEDDING_VECTOR_FILE_NAME
            self.concept_file = model_path / PRETRAINED_EMBEDDING_CONCEPT_FILE_NAME
            self.exists = self.vector_file.exists() and self.concept_file.exists()
        else:
            self.exists = False
        self._initialize_embeddings() if self.exists else self._initialize_empty()

    def _initialize_embeddings(self):
        """Load embeddings and associated concepts from files."""
        self.pretrained_embeddings = np.load(self.vector_file)
        with open(self.concept_file, "rb") as f:
            self.pretrained_concepts = pickle.load(f)

        self.concept_ids = [
            concept["concept_id"] for concept in self.pretrained_concepts
        ]
        self.reverse_concept_id_map = {
            concept_id: i for i, concept_id in enumerate(self.concept_ids)
        }
        self.concept_names = [
            concept["concept_name"] for concept in self.pretrained_concepts
        ]
        self.embed_dim = self.pretrained_embeddings.shape[1]

        assert len(self.pretrained_embeddings) == len(
            self.pretrained_concepts
        ), "The number of embeddings does not match the number of concepts."

    def _initialize_empty(self):
        """Initialize empty attributes for when files do not exist."""
        self.pretrained_embeddings = None
        self.pretrained_concepts = None
        self.concept_ids = None
        self.concept_names = None
        self.reverse_concept_id_map = None
        self.embed_dim = 0

    @property
    def vocab_size(self) -> int:
        """Return the size of the vocabulary."""
        return len(self.pretrained_embeddings) if self.exists else 0

    def is_concept_available(self, concept_id: str) -> bool:
        """Check if a given concept ID is available."""
        return self.exists and concept_id in self.concept_ids

    def get_concept_embeddings(self, concept_id: str) -> Optional[np.ndarray]:
        """
        Retrieve the embedding vector for a given concept ID.

        Returns None if the concept ID is not available.
        """
        if self.is_concept_available(concept_id):
            return self.pretrained_embeddings[self.reverse_concept_id_map[concept_id]]
        return None

    def save(self, model_folder: str):
        """Save the embeddings and concepts to the specified folder."""
        if self.exists:
            model_path = Path(model_folder)
            np.save(
                model_path / PRETRAINED_EMBEDDING_VECTOR_FILE_NAME,
                self.pretrained_embeddings,
            )
            with open(model_path / PRETRAINED_EMBEDDING_CONCEPT_FILE_NAME, "wb") as f:
                pickle.dump(self.pretrained_concepts, f)
